/*************************************************************************
 * chamfer.c - chamfer distance mask computation
 *
 * $$
 *
 * LICENSE:
 * GPL v3.0 (see gpl-3.0.txt for details)
 *
 * AUTHOR:
 * Gregoire Malandain (gregoire.malandain@inria.fr)
 * 
 * CREATION DATE: 
 * May, 10 2005
 *
 * ADDITIONS, CHANGES
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <sys/types.h>
#include <sys/times.h>

#include <chamfer.h>










/*==================================================
 *
 * MISC
 *
 ==================================================*/



static int _debug_ = 0;

void _set_chamfer_debug() 
{
  _debug_ = 1;
}

void _unset_chamfer_debug() 
{
  _debug_ = 0;
}


static int _trace_ = 0;

void _set_chamfer_trace() 
{
  _trace_ = 1;
}

void _unset_chamfer_trace() 
{
  _trace_ = 0;
}


static int _verbose_ = 0;
void _set_chamfer_verbose( int v ) 
{
  if ( v >= 0 )
    _verbose_ = v;
}


static int _alloc_ = 100;
static int _be_strict_ = 0;
static double _tiny_ = 1e-12;



CriteriumType _criterium_ = _CONVEXITY_;

void _set_criterium_to_maximality()
{
  _criterium_ = _MAXIMALITY_;
}

void _set_criterium_to_convexity()
{
  _criterium_ = _CONVEXITY_;
}









/*==================================================
 *
 * VECTORS
 *
 ==================================================*/



void _init_vector( Vector *v ) 
{
  v->x = 0;
  v->y = 0;
  v->z = 0;

  v->vx = 0.0;
  v->vy = 0.0;
  v->vz = 0.0;
  
  v->vn = 0.0;  

  v->ux = 0.0;
  v->uy = 0.0;
  v->uz = 0.0;
  
  v->c = NULL;
  v->n_cones = 0;
}



void _init_vector_list( VectorList *vl )
{
  vl->allocatedVectors = NULL;
  vl->n_allocated_vectors = 0;
  vl->n_vectors = 0;
}



void _free_vector_list( VectorList *vl )
{
  int i;
  
  for ( i=0; i<vl->n_allocated_vectors; i++ ) {
    if ( vl->allocatedVectors[i].c != NULL )
      free( vl->allocatedVectors[i].c );
  }

  if ( vl->allocatedVectors != NULL ) free( vl->allocatedVectors );
  _init_vector_list( vl );
}



void _compute_2D_uv_vectors( Vector *v, double *s ) 
{
  v->vx = s[0] * v->x;
  v->vy = s[1] * v->y;
  v->vn = sqrt( v->vx*v->vx + v->vy*v->vy );
  v->ux = v->vx / v->vn;
  v->uy = v->vy / v->vn;
}



void _compute_3D_uv_vectors( Vector *v, double *s ) 
{
  v->vx = s[0] * v->x;
  v->vy = s[1] * v->y;
  v->vz = s[2] * v->z;
  v->vn = sqrt( v->vx*v->vx + v->vy*v->vy + v->vz*v->vz );
  v->ux = v->vx / v->vn;
  v->uy = v->vy / v->vn;
  v->uz = v->vz / v->vn;
}



void _print_2D_vector( FILE *f, Vector *v )
{
  fprintf( f, "(%2d %2d)", v->x, v->y );
}



void _print_2D_full_vector( FILE *f, Vector *v )
{
  fprintf( f, "   " );
  fprintf( f, "(x,y)=(%2d %2d) ", v->x, v->y );
  fprintf( f, "v=(%f %f) ", v->vx, v->vy );
  fprintf( f, "|v|=%f ", v->vn );
  fprintf( f, "flag=%d ", v->flag );
  fprintf( f, "(w,f,l)=(%d %d,%d) ", v->w, v->f, v->l );
}



void _print_3D_vector( FILE *f, Vector *v )
{
  fprintf( f, "(%2d %2d %2d)", v->x, v->y, v->z );
}



int _add_vector_to_list( VectorList *vl, int x, int y, int z, double *s )
{
  char *proc = "_add_vector_to_list";
  Vector *newv;
  int i, nalloc;
  
  /* look for the vector
   */
  for ( i=0; i<vl->n_vectors; i++ ) {
    if ( vl->allocatedVectors[i].x == x 
	 && vl->allocatedVectors[i].y == y 
	 && vl->allocatedVectors[i].z == z ) {
      return ( i );
    }
  }

  /* is allocated space sufficient ?
   */
  if ( vl->n_vectors >= vl->n_allocated_vectors ) {

    if ( vl->n_allocated_vectors <= 0 ) 
      nalloc = _alloc_;
    else
      nalloc = vl->n_allocated_vectors + _alloc_;

    /* allocation
     */
    newv = (Vector*)malloc( nalloc * sizeof( Vector ) );
    if ( newv == NULL ) {
      fprintf( stderr, "%s: allocation error\n", proc );
      exit( -1 );
    }
    
    /* copy and/or initialization
     */
    if ( vl->n_allocated_vectors <= 0 ) {
      for ( i=0; i<nalloc; i++ ) {
	_init_vector( &(newv[i]) );
      }
    }
    else {
      memcpy( newv, vl->allocatedVectors, 
	      vl->n_allocated_vectors * sizeof( Vector ) );
      for ( i=vl->n_vectors; i<nalloc; i++ ) {
	_init_vector( &(newv[i]) );
      }
      free( vl->allocatedVectors );
    }

    vl->allocatedVectors = newv;
    vl->n_allocated_vectors = nalloc;
  }
  

  /* create the vector
   */
  vl->n_vectors ++;
  vl->allocatedVectors[ vl->n_vectors-1 ].x = x;
  vl->allocatedVectors[ vl->n_vectors-1 ].y = y;
  vl->allocatedVectors[ vl->n_vectors-1 ].z = z;
  if ( z == 0 )
    _compute_2D_uv_vectors( &(vl->allocatedVectors[ vl->n_vectors-1 ]), s );
  else
    _compute_3D_uv_vectors( &(vl->allocatedVectors[ vl->n_vectors-1 ]), s );
  return( vl->n_vectors-1 );

}


double _norme1v_2D( Vector *v )
{
  return( v->vx + v->vy );
}



double _normeIv_2D( Vector *v )
{
  if ( v->vx > v->vy )
    return( v->vx );
  else
    return( v->vy );
}



double _norme2v_2D( Vector *v )
{
  return( sqrt( v->vx * v->vx + v->vy * v->vy ) );
}



double _norme1v_3D( Vector *v )
{
  return( v->vx + v->vy  + v->vz );
}



double _normeIv_3D( Vector *v )
{
  if ( v->vx > v->vy ) {
    if ( v->vx > v->vz )
      return( v->vx );
    else
      return( v->vz );
  }
  else {
    if ( v->vy > v->vz )
      return( v->vy );
    else
      return( v->vz );
  }
}



double _norme2v_3D( Vector *v )
{
  return( sqrt( v->vx * v->vx + v->vy * v->vy + v->vz * v->vz ) );
}













/*==================================================
 *
 * CONES
 *
 ==================================================*/



void _init_cone( Cone *c )
{
  c->iv1 = -1;
  c->iv2 = -1;
  c->iv3 = -1;
  
  c->type = _UNSET_;

  c->error  = 1.0;
}



void _init_cone_list( ConeList *cl )
{
  cl->allocatedCones = NULL;
  cl->n_allocated_cones = 0;
  cl->n_cones = 0;
}



int _alloc_cone_list(  ConeList *cl, int n )
{
  int i;

  if ( n < 1 ) return( -1 );
  cl->allocatedCones = (Cone *)malloc( n * sizeof( Cone ) );
  if ( cl->allocatedCones == NULL ) {
    return( -1 );
  }

  for ( i=0; i<n; i++ ) {
    _init_cone( &(cl->allocatedCones[i]) );
  }

  cl->n_allocated_cones = n;
  cl->n_cones = 0;
  
  return( 1 );
}



void _free_cone_list( ConeList *cl )
{
  free( cl->allocatedCones );
  _init_cone_list( cl );
}



void _sort_cones( Cone *cl, int left, int right )
{
  int i, last;
  Cone c;

  if ( left >= right ) return;

  c = cl[left];   cl[left] = cl[(left+right)/2];   cl[(left+right)/2] = c;
  
  last = left;

  for ( i = left+1; i <= right; i++ )
    if ( cl[i].error > cl[left].error ) {
      c = cl[++last];   cl[last] = cl[i];   cl[i] = c;
    }

  c = cl[left];   cl[left] = cl[last];   cl[last] = c;

  _sort_cones( cl, left, last-1 );
  _sort_cones( cl, last+1, right );
  
}


void _move_cone_backward( ConeList *cl, int n )
{
  int i=n;
  Cone c;
  
  while ( i > 0 
	  && cl->allocatedCones[i].error > cl->allocatedCones[i-1].error ) {
    c = cl->allocatedCones[i];
    cl->allocatedCones[i] = cl->allocatedCones[i-1];
    cl->allocatedCones[i-1] = c;
    i --;
  }
}



void _move_cone_forward( ConeList *cl, int n )
{
  int i=n;
  Cone c;
  
  while ( i < cl->n_cones-1 
	  && cl->allocatedCones[i].error < cl->allocatedCones[i+1].error ) {
    c = cl->allocatedCones[i];
    cl->allocatedCones[i] = cl->allocatedCones[i+1];
    cl->allocatedCones[i+1] = c;
    i ++;
  }
}



void _compute_2D_cone_error( Cone *c, Vector *v1, Vector *v2 )
{
  double x, y, h;

  x = v1->ux + v2->ux;
  y = v1->uy + v2->uy;

  h = sqrt( x*x + y*y )/2;

  c->type = _INSIDE_;
  c->error = (1-h)/(1+h);
}



void _compute_3D_cone_error( Cone *c, Vector *v1, Vector *v2, Vector *v3 )
{
  char *proc = "_compute_3D_cone_error";
  double sp12, sp23, sp31;
  double sign12, sign23, sign31;
  int nn = 0;
  double h, n, x, y, z;

  sp12 = v1->ux * v2->ux + v1->uy * v2->uy + v1->uz * v2->uz;

  sp23 = v2->ux * v3->ux + v2->uy * v3->uy + v2->uz * v3->uz;

  sp31 = v3->ux * v1->ux + v3->uy * v1->uy + v3->uz * v1->uz;

  sign12 = 1.0 + sp12 - sp23 - sp31;
  sign23 = 1.0 - sp12 + sp23 - sp31;
  sign31 = 1.0 - sp12 - sp23 + sp31;

  if ( sign12 < 0 ) nn++;
  if ( sign23 < 0 ) nn++;
  if ( sign31 < 0 ) nn++;

  if ( nn >= 2 ) {
    fprintf( stderr, "%s: two negative scalar products !?!\n", proc );
    fprintf( stderr, "    this should never occur\n" );
    exit( 0 );
  }

  if ( nn == 0 ) {
    c->type = _INSIDE_;

    n = v1->ux * v2->uy * v3->uz - v1->ux * v2->uz * v3->uy
      + v1->uy * v2->uz * v3->ux - v1->uy * v2->ux * v3->uz
      + v1->uz * v2->ux * v3->uy - v1->uz * v2->uy * v3->ux;

    x = v2->uy * v3->uz - v2->uz * v3->uy + v3->uy * v1->uz 
      - v3->uz * v1->uy + v1->uy * v2->uz - v1->uz * v2->uy;
    
    y = v2->uz * v3->ux - v2->ux * v3->uz + v3->uz * v1->ux 
      - v3->ux * v1->uz + v1->uz * v2->ux - v1->ux * v2->uz;
    
    z = v2->ux * v3->uy - v2->uy * v3->ux + v3->ux * v1->uy 
      - v3->uy * v1->ux + v1->ux * v2->uy - v1->uy * v2->ux;

    h = n /sqrt( x*x + y*y + z*z );
    c->error = (1-h)/(1+h);
    return; 
  }

  if ( sign12 < 0 ) {
    c->type = _OUT_IJ_;

    x = v1->ux + v2->ux;
    y = v1->uy + v2->uy;
    z = v1->uz + v2->uz;

    h = sqrt( x*x + y*y + z*z )/2;
    c->error = (1-h)/(1+h);
    return;
  }

  if ( sign23 < 0 ) {
    c->type = _OUT_JK_;

    x = v2->ux + v3->ux;
    y = v2->uy + v3->uy;
    z = v2->uz + v3->uz;

    h = sqrt( x*x + y*y + z*z )/2;
    c->error = (1-h)/(1+h);
    return;
  }

  if ( sign31 < 0 ) {
    c->type = _OUT_KI_;

    x = v3->ux + v1->ux;
    y = v3->uy + v1->uy;
    z = v3->uz + v1->uz;

    h = sqrt( x*x + y*y + z*z )/2;
    c->error = (1-h)/(1+h);
    return;
  }

  fprintf( stderr, "%s: unexpected end\n", proc );
  c->error = 1.0;
  return;
}



int _add_cone_to_list( VectorList *vl, ConeList *cl, int iv1, int iv2, int iv3 )
{
  char *proc = "_add_cone_to_list";
  Cone *newc;
  int i, nalloc;
  
  /* is allocated space sufficient ?
   */
  if ( cl->n_cones >= cl->n_allocated_cones ) {

    if ( cl->n_allocated_cones <= 0 ) 
      nalloc = _alloc_;
    else
      nalloc = cl->n_allocated_cones + _alloc_;

    /* allocation
     */
    newc = (Cone*)malloc( nalloc * sizeof( Cone ) );
    if ( newc == NULL ) {
      fprintf( stderr, "%s: allocation error\n", proc );
      exit( -1 );
    }
    
    /* copy and/or initialization
     */
    if ( cl->n_allocated_cones <= 0 ) {
      for ( i=0; i<nalloc; i++ ) {
	_init_cone( &(newc[i]) );
      }
    }
    else {
      memcpy( newc, cl->allocatedCones, 
	      cl->n_allocated_cones * sizeof( Cone ) );
      for ( i=cl->n_cones; i<nalloc; i++ ) {
	_init_cone( &(newc[i]) );
      }
      free( cl->allocatedCones );
    }

    cl->allocatedCones = newc;
    cl->n_allocated_cones = nalloc;
  }


  /* create the cone
   */
  cl->n_cones ++;
  cl->allocatedCones[ cl->n_cones-1 ].iv1 = iv1;
  cl->allocatedCones[ cl->n_cones-1 ].iv2 = iv2;
  cl->allocatedCones[ cl->n_cones-1 ].iv3 = iv3;
  if ( iv3 < 0 )
    _compute_2D_cone_error( &(cl->allocatedCones[cl->n_cones-1]),
			    &(vl->allocatedVectors[cl->allocatedCones[cl->n_cones-1].iv1]),
			    &(vl->allocatedVectors[cl->allocatedCones[cl->n_cones-1].iv2]) );
  else
    _compute_3D_cone_error( &(cl->allocatedCones[ cl->n_cones-1 ]),
			    &(vl->allocatedVectors[cl->allocatedCones[cl->n_cones-1].iv1]),
			    &(vl->allocatedVectors[cl->allocatedCones[cl->n_cones-1].iv2]),
			    &(vl->allocatedVectors[cl->allocatedCones[cl->n_cones-1].iv3]) );
  return( cl->n_cones-1 );

}



int _check_2D_maximality( VectorList *vl, Cone *c )
{
  double sp;

  sp = vl->allocatedVectors[c->iv1].vx * vl->allocatedVectors[c->iv2].vx 
    + vl->allocatedVectors[c->iv1].vy * vl->allocatedVectors[c->iv2].vy;

  if ( vl->allocatedVectors[c->iv1].vn * vl->allocatedVectors[c->iv1].vn 
       * vl->allocatedVectors[c->iv2].w >= sp * vl->allocatedVectors[c->iv1].w 
       && vl->allocatedVectors[c->iv2].vn * vl->allocatedVectors[c->iv2].vn 
       * vl->allocatedVectors[c->iv1].w >= sp * vl->allocatedVectors[c->iv2].w )
    return( 1 );

  return( 0 );
}


int _3x3_determinant( int mat[3][3] )
{
  return( mat[0][0]*mat[1][1]*mat[2][2]
	  + mat[0][1]*mat[1][2]*mat[2][0]
	  + mat[0][2]*mat[1][0]*mat[2][1]
	  - mat[2][0]*mat[1][1]*mat[0][2]
	  - mat[2][1]*mat[1][2]*mat[0][0]
	  - mat[2][2]*mat[1][0]*mat[0][1] );
}



int _check_2D_convexity( VectorList *vl, ConeList *cl, Cone *c )
{
  char *proc="_check_2D_convexity";
  Vector *v1, *v2;
  int i;
  int mat[3][3];
  Cone *cc;
  int convexity_1 = -2;
  int convexity_2 = -2;
  int sp, n1, n2;
  Vector sym;


  v1 = &(vl->allocatedVectors[ c->iv1 ]);
  v2 = &(vl->allocatedVectors[ c->iv2 ]);

  if ( v1->flag != 1 || v2->flag != 1 )
    return( -1 );

  /* convexity at side iv1
   */
  _init_vector( &sym );
  mat[1][0] = v1->x;
  mat[1][1] = v1->y;
  mat[1][2] = v1->w;
  mat[2][0] = v2->x;
  mat[2][1] = v2->y;
  mat[2][2] = v2->w;

  for ( i=0; i<cl->n_cones && convexity_1 == -2; i++ ) {
    cc = &(cl->allocatedCones[i]);
    if ( cc->iv2 != c->iv1 ) continue;
    if ( vl->allocatedVectors[cc->iv1].flag != 1 ) {
      convexity_1 = -1;
      continue;
    }
    mat[0][0] = vl->allocatedVectors[cc->iv1].x;
    mat[0][1] = vl->allocatedVectors[cc->iv1].y;
    mat[0][2] = vl->allocatedVectors[cc->iv1].w;
    if ( _3x3_determinant( mat ) < 0 ) {
      return( 0 );
    }
    convexity_1 = 1;
  }

  if ( convexity_1 == -2 ) {
    n1 = v1->x * v1->x + v1->y * v1->y; 
    sp = v1->x * v2->x + v1->y * v2->y; 
    sym.x = 2 * sp * v1->x / n1 - v2->x;
    sym.y = 2 * sp * v1->y / n1 - v2->y;
    if ( v1->x != 1 || v1->y != 0 ) {
      fprintf( stderr, "%s: unexpected symmetry (#1) w.r.t. (%d %d)\n", proc,
	       v1->x, v1->y );
    }
    mat[0][0] = sym.x;
    mat[0][1] = sym.y;
    mat[0][2] = v2->w;
    if ( _3x3_determinant( mat ) < 0 ) {
      return( 0 );
    }
    convexity_1 = 1;
  }


  /* convexity at side iv2
   */
  _init_vector( &sym );
  mat[0][0] = v1->x;
  mat[0][1] = v1->y;
  mat[0][2] = v1->w;
  mat[1][0] = v2->x;
  mat[1][1] = v2->y;
  mat[1][2] = v2->w;

  for ( i=0; i<cl->n_cones && convexity_2 == -2; i++ ) {
    cc = &(cl->allocatedCones[i]);
    if ( cc->iv1 != c->iv2 ) continue;
    if ( vl->allocatedVectors[cc->iv2].flag != 1 ) {
      convexity_2 = -1;
      continue;
    }
    mat[2][0] = vl->allocatedVectors[cc->iv2].x;
    mat[2][1] = vl->allocatedVectors[cc->iv2].y;
    mat[2][2] = vl->allocatedVectors[cc->iv2].w;
    if ( _3x3_determinant( mat ) < 0 ) {
      return( 0 );
    }
    convexity_2 = 1;
  }

  if ( convexity_2 == -2 ) {
    n2 = v2->x * v2->x + v2->y * v2->y; 
    sp = v1->x * v2->x + v1->y * v2->y; 
    sym.x = 2 * sp * v2->x / n2 - v1->x;
    sym.y = 2 * sp * v2->y / n2 - v1->y;
    if ( (v2->x != 1 || v2->y != 1) && (v2->x != 0 || v2->y != 1) ) {
      fprintf( stderr, "%s: unexpected symmetry (#2) w.r.t. (%d %d)\n", proc,
	       v2->x, v2->y );
    }
    mat[2][0] = sym.x;
    mat[2][1] = sym.y;
    mat[2][2] = v1->w;
    if ( _3x3_determinant( mat ) < 0 ) {
      return( 0 );
    }
    convexity_2 = 1;
  }

  return( 1 );

}






void _compute_2D_cone_max( Vector *v1, Vector *v2, 
			   double *dmax )
{
  char *proc = "_compute_2D_cone_max";
  double sp;
  double x, y;

  sp = v1->vx * v2->vx + v1->vy * v2->vy;

  if ( v1->w * v2->vn * v2->vn >= sp *v2->w ) {
    if ( v2->w * v1->vn * v1->vn >= sp *v1->w ) {
      x = v1->w * v2->vx - v2->w * v1->vx;
      y = v1->w * v2->vy - v2->w * v1->vy;
      *dmax = sqrt( x*x + y*y ) / fabs( v1->vx * v2->vy - v1->vy * v2->vx );
    }
    else {
      *dmax = v1->w / v1->vn;
    }
  }
  else {
    if ( v2->w * v1->vn * v1->vn >= sp *v1->w ) {
      *dmax = v2->w / v2->vn;
    }
    else {
      fprintf( stdout, "%s: this case should never occur\n", proc );
    }
  }
}



void _compute_2D_cone_min_max( Vector *v1, Vector *v2, 
			  double *dmin, double *dmax )
{
  *dmin = v1->w / v1->vn; 
  if ( *dmin > v2->w / v2->vn ) *dmin = v2->w / v2->vn;

 _compute_2D_cone_max( v1, v2, dmax );
}



void _compute_2D_mask_min_max( VectorList *vl, ConeList *cl, 
			       double *dmin, double *dmax )
{
  int i;
  double ldmax;
  
  *dmin = vl->allocatedVectors[0].w / vl->allocatedVectors[0].vn;
  
  for ( i=1; i<vl->n_vectors; i++ ) 
    if ( *dmin > vl->allocatedVectors[i].w / vl->allocatedVectors[i].vn )
      *dmin = vl->allocatedVectors[i].w / vl->allocatedVectors[i].vn;

  *dmax = *dmin;
  for ( i=0; i<cl->n_cones; i++ ) {
    _compute_2D_cone_max( &vl->allocatedVectors[cl->allocatedCones[i].iv1],
			  &vl->allocatedVectors[cl->allocatedCones[i].iv2],
			  &ldmax );
    if ( *dmax < ldmax ) *dmax = ldmax;
  }
}


int _check_3D_maximality( VectorList *vl, Cone *c )
{
  Vector *v1, *v2, *v3;
  double u12[3], u23[3], u31[3], u[3];
  

  v1 = &(vl->allocatedVectors[ c->iv1 ]);
  v2 = &(vl->allocatedVectors[ c->iv2 ]);
  v3 = &(vl->allocatedVectors[ c->iv3 ]);

  u23[0] = v2->vy * v3->vz - v2->vz * v3->vy;
  u23[1] = v2->vz * v3->vx - v2->vx * v3->vz;
  u23[2] = v2->vx * v3->vy - v2->vy * v3->vx;

  u31[0] = v3->vy * v1->vz - v3->vz * v1->vy;
  u31[1] = v3->vz * v1->vx - v3->vx * v1->vz;
  u31[2] = v3->vx * v1->vy - v3->vy * v1->vx;

  u12[0] = v1->vy * v2->vz - v1->vz * v2->vy;
  u12[1] = v1->vz * v2->vx - v1->vx * v2->vz;
  u12[2] = v1->vx * v2->vy - v1->vy * v2->vx;

  u[0] = v1->w * u23[0] + v2->w * u31[0] + v3->w * u12[0];
  u[1] = v1->w * u23[1] + v2->w * u31[1] + v3->w * u12[1];
  u[2] = v1->w * u23[2] + v2->w * u31[2] + v3->w * u12[2];

  if ( _verbose_ >= 3  ) {
    fprintf( stdout, "CONVEXITY " ); 
    switch ( c->type ) {
    case _UNSET_  : fprintf( stdout, "--- " ); break;
    case _INSIDE_ : fprintf( stdout, "IJK " ); break;
    case _OUT_IJ_ : fprintf( stdout, "IJ- " ); break;
    case _OUT_JK_ : fprintf( stdout, "-JK " ); break;
    case _OUT_KI_ : fprintf( stdout, "I-K " ); break;
    }
    fprintf( stdout, "n1=(%d %d %d) w=%d  n2=(%d %d %d)  w=%d  n3=(%d %d %d) w=%d  \n",
	     v1->x, v1->y, v1->z, v1->w, 
	     v2->x, v2->y, v2->z, v2->w, 
	     v3->x, v3->y, v3->z, v3->w );
    fprintf( stdout, "          v1 x v2 = (%f %f %f)\n", u12[0], u12[1], u12[2] );
    fprintf( stdout, "          v2 x v3 = (%f %f %f)\n", u23[0], u23[1], u23[2] );
    fprintf( stdout, "          v3 x v1 = (%f %f %f)\n", u31[0], u31[1], u31[2] );
    fprintf( stdout, "          w1 v23 + w2 v31 + w3 v12 = (%f %f %f)\n",
	     u[0], u[1], u[2] );
    fprintf( stdout, "          C#1 = U x v23 = %f\n",
	     u[0]*u23[0] + u[1]*u23[1] + u[2]*u23[2] );
    fprintf( stdout, "          C#1 = U x v31 = %f\n",
	     u[0]*u31[0] + u[1]*u31[1] + u[2]*u31[2] );
    fprintf( stdout, "          C#1 = U x v12 = %f\n",
	     u[0]*u12[0] + u[1]*u12[1] + u[2]*u12[2] );
  }

  if ( _be_strict_ == 0 ) {
    if ( (u[0]*u23[0] + u[1]*u23[1] + u[2]*u23[2] >= 0 
	  || fabs (u[0]*u23[0] + u[1]*u23[1] + u[2]*u23[2]) < _tiny_)
	 && (u[0]*u31[0] + u[1]*u31[1] + u[2]*u31[2] >= 0 
	     || fabs (u[0]*u31[0] + u[1]*u31[1] + u[2]*u31[2]) < _tiny_)
	 && (u[0]*u12[0] + u[1]*u12[1] + u[2]*u12[2] >= 0 
	     || fabs (u[0]*u12[0] + u[1]*u12[1] + u[2]*u12[2]) < _tiny_) ) {
	       
	       return( 1 );
	     }
  }
  else {
    if ( u[0]*u23[0] + u[1]*u23[1] + u[2]*u23[2] >= 0 
	 && u[0]*u31[0] + u[1]*u31[1] + u[2]*u31[2] >= 0 
	 && u[0]*u12[0] + u[1]*u12[1] + u[2]*u12[2] >= 0 ) {

	   return( 1 );
	 }
  }
  return( 0 );
}



int _search_adjacent_3D_cone( ConeList *cl, int iv1, int iv2, int iv3, int *side )
{
  int i;
  Cone *cc;
  
  *side = -1;
  
  for ( i=0; i<cl->n_cones; i++ ) {
    cc = &(cl->allocatedCones[i]);
    /* IJ case
     */
    if ( (cc->iv1 == iv1 && cc->iv2 == iv2 && cc->iv3 != iv3) 
	 || (cc->iv2 == iv1 && cc->iv1 == iv2 && cc->iv3 != iv3) ) {
      *side = 2;
      return( i );
    }
    /* JK case
     */
    if ( (cc->iv2 == iv1 && cc->iv3 == iv2 && cc->iv1 != iv3) 
	 || (cc->iv3 == iv1 && cc->iv2 == iv2 && cc->iv1 != iv3) ) {
      *side = 0;
      return( i );
    }
    /* KI case
     */
    if ( (cc->iv3 == iv1 && cc->iv1 == iv2 && cc->iv2 != iv3) 
	 || (cc->iv1 == iv1 && cc->iv3 == iv2 && cc->iv2 != iv3) ) {
      *side = 1;
      return( i );
    }
  }
  return( -1 );
}



int _4x4_determinant( int mat[4][4] )
{
  if ( _debug_ ) {
    fprintf( stdout, "\n" );
    fprintf( stdout, "[ %3d %3d %3d %3d ]\n",
	     mat[0][0], mat[1][0], mat[2][0], mat[3][0] );
    fprintf( stdout, "[ %3d %3d %3d %3d ]\n",
	     mat[0][1], mat[1][1], mat[2][1], mat[3][1] );
    fprintf( stdout, "[ %3d %3d %3d %3d ]\n",
	     mat[0][2], mat[1][2], mat[2][2], mat[3][2] );
    fprintf( stdout, "[ %3d %3d %3d %3d ]\n",
	     mat[0][3], mat[1][3], mat[2][3], mat[3][3] );
    fprintf( stdout, "\n" );
  }

  return( mat[0][0] * ( mat[1][1]*mat[2][2]*mat[3][3]
			+ mat[1][2]*mat[2][3]*mat[3][1]
			+ mat[1][3]*mat[2][1]*mat[3][2]
			- mat[3][1]*mat[2][2]*mat[1][3]
			- mat[3][2]*mat[2][3]*mat[1][1]
			- mat[3][3]*mat[2][1]*mat[1][2] )
	  - mat[1][0] * ( mat[0][1]*mat[2][2]*mat[3][3]
			+ mat[0][2]*mat[2][3]*mat[3][1]
			+ mat[0][3]*mat[2][1]*mat[3][2]
			- mat[3][1]*mat[2][2]*mat[0][3]
			- mat[3][2]*mat[2][3]*mat[0][1]
			- mat[3][3]*mat[2][1]*mat[0][2] )
	  + mat[2][0] * ( mat[0][1]*mat[1][2]*mat[3][3]
			+ mat[0][2]*mat[1][3]*mat[3][1]
			+ mat[0][3]*mat[1][1]*mat[3][2]
			- mat[3][1]*mat[1][2]*mat[0][3]
			- mat[3][2]*mat[1][3]*mat[0][1]
			- mat[3][3]*mat[1][1]*mat[0][2] )
	  - mat[3][0] * ( mat[0][1]*mat[1][2]*mat[2][3]
			+ mat[0][2]*mat[1][3]*mat[2][1]
			+ mat[0][3]*mat[1][1]*mat[2][2]
			- mat[2][1]*mat[1][2]*mat[0][3]
			- mat[2][2]*mat[1][3]*mat[0][1]
			- mat[2][3]*mat[1][1]*mat[0][2] ) );
}



int _check_3D_convexity( VectorList *vl, ConeList *cl, Cone *c )
{
  Vector *v1, *v2, *v3, *v4 = NULL;
  int adjcone, adjside;
  Vector sym, vp;
  int n, sp;
  int mat[4][4];

  v1 = &(vl->allocatedVectors[ c->iv1 ]);
  v2 = &(vl->allocatedVectors[ c->iv2 ]);
  v3 = &(vl->allocatedVectors[ c->iv3 ]);

  if ( v1->flag != 1 || v2->flag != 1 || v3->flag != 1 )
    return( -1 );

  /* check convexity for JK 
   */
  mat[0][0] = v1->x; mat[0][1] = v1->y; mat[0][2] = v1->z; mat[0][3] = v1->w;
  mat[1][0] = v2->x; mat[1][1] = v2->y; mat[1][2] = v2->z; mat[1][3] = v2->w;
  mat[2][0] = v3->x; mat[2][1] = v3->y; mat[2][2] = v3->z; mat[2][3] = v3->w;
  adjcone = _search_adjacent_3D_cone( cl, c->iv2, c->iv3, c->iv1, &adjside );
  if ( adjcone >= 0 ) {
    switch ( adjside ) {
    case 2 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv3 ]);
      break;
    case 0 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv1 ]);
      break;
    case 1 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv2 ]);
      break;
    }
    if ( v4->flag == 1 ) {
      mat[3][0] = v4->x; mat[3][1] = v4->y; mat[3][2] = v4->z; mat[3][3] = v4->w;
      if ( _4x4_determinant( mat ) < 0 ) {
	if ( _debug_ ) 
	  fprintf( stdout, "convexity: unsuccessful convexity test (JK)\n" );
	return( 0 );
      }
      if ( _debug_ ) 
	fprintf( stdout, "convexity: successful convexity test (JK)\n" );
    }
    else {
      if ( _debug_ ) 
	fprintf( stdout, "convexity: unchecked convexity test (JK)\n" );
    }
  }
  else {
    /* symmetrise cone
     */
    vp.x = v2->y * v3->z - v2->z * v3->y;
    vp.y = v2->z * v3->x - v2->x * v3->z;
    vp.z = v2->x * v3->y - v2->y * v3->x;
    n = vp.x * vp.x + vp.y * vp.y + vp.z * vp.z;
    sp = vp.x * v1->x + vp.y * v1->y + vp.z * v1->z;
    sym.x = v1->x - 2 * sp * vp.x / n;
    sym.y = v1->y - 2 * sp * vp.y / n;
    sym.z = v1->z - 2 * sp * vp.z / n;
    if ( _debug_ ) {
      fprintf( stdout, 
	       "SYM (#1) of (%d %d %d) wrt (%d %d %d),(%d %d %d) is (%d %d %d)\n",
	       v1->x, v1->y, v1->z,
	       v2->x, v2->y, v2->z,
	       v3->x, v3->y, v3->z,
	       sym.x, sym.y, sym.z );
    }
    mat[3][0] = sym.x; mat[3][1] = sym.y; mat[3][2] = sym.z; mat[3][3] = v1->w;
    if ( _4x4_determinant( mat ) < 0 ) {
	if ( _debug_ ) 
	  fprintf( stdout, "convexity: unsuccessful convexity symmetry test (JK)\n" );
	return( 0 );
    }
    if ( _debug_ ) 
      fprintf( stdout, "convexity: successful convexity symmetry test (JK)\n" );
  }
  
  /* check convexity for KI 
   */
  mat[0][0] = v2->x; mat[0][1] = v2->y; mat[0][2] = v2->z; mat[0][3] = v2->w;
  mat[1][0] = v3->x; mat[1][1] = v3->y; mat[1][2] = v3->z; mat[1][3] = v3->w;
  mat[2][0] = v1->x; mat[2][1] = v1->y; mat[2][2] = v1->z; mat[2][3] = v1->w;
  adjcone = _search_adjacent_3D_cone( cl, c->iv3, c->iv1, c->iv2, &adjside );
  if ( adjcone >= 0 ) {
    switch ( adjside ) {
    case 2 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv3 ]);
      break;
    case 0 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv1 ]);
      break;
    case 1 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv2 ]);
      break;
    }
    if ( v4->flag == 1 ) {
      mat[3][0] = v4->x; mat[3][1] = v4->y; mat[3][2] = v4->z; mat[3][3] = v4->w;
      if ( _4x4_determinant( mat ) < 0 ) {
	if ( _debug_ ) 
	  fprintf( stdout, "convexity: unsuccessful convexity test (KI)\n" );
	return( 0 );
      }
      if ( _debug_ ) 
	fprintf( stdout, "convexity: successful convexity test (KI)\n" );
    }
    else {
      if ( _debug_ ) 
	fprintf( stdout, "convexity: unchecked convexity test (KI)\n" );
    }
  }
  else {
    /* symmetrise cone
     */
    vp.x = v3->y * v1->z - v3->z * v1->y;
    vp.y = v3->z * v1->x - v3->x * v1->z;
    vp.z = v3->x * v1->y - v3->y * v1->x;
    n = vp.x * vp.x + vp.y * vp.y + vp.z * vp.z;
    sp = vp.x * v2->x + vp.y * v2->y + vp.z * v2->z;
    sym.x = v2->x - 2 * sp * vp.x / n;
    sym.y = v2->y - 2 * sp * vp.y / n;
    sym.z = v2->z - 2 * sp * vp.z / n;
    if ( _debug_ ) {
      fprintf( stdout, 
	       "SYM (#2) of (%d %d %d) wrt (%d %d %d),(%d %d %d) is (%d %d %d)\n",
	       v2->x, v2->y, v2->z,
	       v3->x, v3->y, v3->z,
	       v1->x, v1->y, v1->z,
	       sym.x, sym.y, sym.z );
    }
    mat[3][0] = sym.x; mat[3][1] = sym.y; mat[3][2] = sym.z; mat[3][3] = v2->w;
    if ( _4x4_determinant( mat ) < 0 ) {
      if ( _debug_ ) 
	fprintf( stdout, "convexity: unsuccessful convexity symmetry test (KI)\n" );
      return( 0 );
    }
    if ( _debug_ ) 
      fprintf( stdout, "convexity: successful convexity symmetry test (KI)\n" );
  }
  
  /* check convexity for IJ
   */
  mat[0][0] = v3->x; mat[0][1] = v3->y; mat[0][2] = v3->z; mat[0][3] = v3->w;
  mat[1][0] = v1->x; mat[1][1] = v1->y; mat[1][2] = v1->z; mat[1][3] = v1->w;
  mat[2][0] = v2->x; mat[2][1] = v2->y; mat[2][2] = v2->z; mat[2][3] = v2->w;
  adjcone = _search_adjacent_3D_cone( cl, c->iv1, c->iv2, c->iv3, &adjside );
  if ( adjcone >= 0 ) {
    switch ( adjside ) {
    case 2 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv3 ]);
      break;
    case 0 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv1 ]);
      break;
    case 1 :
      v4 = &(vl->allocatedVectors[ cl->allocatedCones[adjcone].iv2 ]);
      break;
    }
    if ( v4->flag == 1 ) {
      mat[3][0] = v4->x; mat[3][1] = v4->y; mat[3][2] = v4->z; mat[3][3] = v4->w;
      if ( _4x4_determinant( mat ) < 0 ) {
	if ( _debug_ ) 
	  fprintf( stdout, "convexity: unsuccessful convexity test (IJ)\n" );
	return( 0 );
      }
      if ( _debug_ ) 
	fprintf( stdout, "convexity: successful convexity test (IJ)\n" );
    }
    else {
      if ( _debug_ ) 
	fprintf( stdout, "convexity: unchecked convexity test (IJ)\n" );
    }
  }
  else {
    /* symmetrise cone
     */
    vp.x = v1->y * v2->z - v1->z * v2->y;
    vp.y = v1->z * v2->x - v1->x * v2->z;
    vp.z = v1->x * v2->y - v1->y * v2->x;
    n = vp.x * vp.x + vp.y * vp.y + vp.z * vp.z;
    sp = vp.x * v3->x + vp.y * v3->y + vp.z * v3->z;
    sym.x = v3->x - 2 * sp * vp.x / n;
    sym.y = v3->y - 2 * sp * vp.y / n;
    sym.z = v3->z - 2 * sp * vp.z / n;
    if ( _debug_ ) {
      fprintf( stdout, 
	       "SYM (#3) of (%d %d %d) wrt (%d %d %d),(%d %d %d) is (%d %d %d)\n",
	       v3->x, v3->y, v3->z,
	       v1->x, v1->y, v1->z,
	       v2->x, v2->y, v2->z,
	       sym.x, sym.y, sym.z );
    }
    mat[3][0] = sym.x; mat[3][1] = sym.y; mat[3][2] = sym.z; mat[3][3] = v3->w;
    if ( _4x4_determinant( mat ) < 0 ) {
	if ( _debug_ ) 
	  fprintf( stdout, "convexity: unsuccessful convexity symmetry test (IJ)\n" );
	return( 0 );
    }
    if ( _debug_ ) 
      fprintf( stdout, "convexity: successful convexity symmetry test (IJ)\n" );
  }
  
  return( 1 );

}


void _compute_3D_cone_max( Vector *v1, Vector *v2, Vector *v3, 
			   double *dmax )
{
  char *proc = "_compute_3D_cone_max";
  double u[3], n;
  double v12[3], v23[3], v31[3], d[3];
  double uv12, uv23, uv31;

  u[0] = v1->w*(v2->vy*v3->vz - v2->vz*v3->vy)
       + v2->w*(v3->vy*v1->vz - v3->vz*v1->vy)
       + v3->w*(v1->vy*v2->vz - v1->vz*v2->vy);
  u[1] = v1->w*(v2->vz*v3->vx - v2->vx*v3->vz)
       + v2->w*(v3->vz*v1->vx - v3->vx*v1->vz)
       + v3->w*(v1->vz*v2->vx - v1->vx*v2->vz);
  u[2] = v1->w*(v2->vx*v3->vy - v2->vy*v3->vx)
       + v2->w*(v3->vx*v1->vy - v3->vy*v1->vx)
       + v3->w*(v1->vx*v2->vy - v1->vy*v2->vx);
	
  v23[0] = v2->vy * v3->vz - v2->vz * v3->vy;
  v23[1] = v2->vz * v3->vx - v2->vx * v3->vz;
  v23[2] = v2->vx * v3->vy - v2->vy * v3->vx;

  v31[0] = v3->vy * v1->vz - v3->vz * v1->vy;
  v31[1] = v3->vz * v1->vx - v3->vx * v1->vz;
  v31[2] = v3->vx * v1->vy - v3->vy * v1->vx;

  v12[0] = v1->vy * v2->vz - v1->vz * v2->vy;
  v12[1] = v1->vz * v2->vx - v1->vx * v2->vz;
  v12[2] = v1->vx * v2->vy - v1->vy * v2->vx;

  uv12 = u[0] * v12[0] + u[1] * v12[1] + u[2] * v12[2];
  uv23 = u[0] * v23[0] + u[1] * v23[1] + u[2] * v23[2];
  uv31 = u[0] * v31[0] + u[1] * v31[1] + u[2] * v31[2];


  if ( uv12 >= 0 ) {
    if ( uv23 >= 0 ) {
      if ( uv31 >= 0 ) {

	*dmax = sqrt( u[0]*u[0] + u[1]*u[1] + u[2]*u[2] );
  
	n = v1->vx * v2->vy * v3->vz
	  + v1->vy * v2->vz * v3->vx
	  + v1->vz * v2->vx * v3->vy
	  - v1->vx * v2->vz * v3->vy
	  - v1->vy * v2->vx * v3->vz
	  - v1->vz * v2->vy * v3->vx;
	*dmax /= n;

      }
      else {
	d[0] = v1->w * v3->vx - v3->w * v1->vx;
	d[1] = v1->w * v3->vy - v3->w * v1->vy;
	d[2] = v1->w * v3->vz - v3->w * v1->vz;
	*dmax = sqrt( (d[0]*d[0]+d[1]*d[1]+d[2]*d[2])
		      / (v31[0]*v31[0]+v31[1]*v31[1]+v31[2]*v31[2]) );
      }
    }
    else {
      if ( uv31 >= 0 ) {
	d[0] = v3->w * v2->vx - v2->w * v3->vx;
	d[1] = v3->w * v2->vy - v2->w * v3->vy;
	d[2] = v3->w * v2->vz - v2->w * v3->vz;
	*dmax = sqrt( (d[0]*d[0]+d[1]*d[1]+d[2]*d[2])
		      / (v23[0]*v23[0]+v23[1]*v23[1]+v23[2]*v23[2]) );
      }
      else {
	*dmax = v3->w / v3->vn;	
      }
    }
  }
  else {
    if ( uv23 >= 0 ) {
      if ( uv31 >= 0 ) {
	d[0] = v2->w * v1->vx - v1->w * v2->vx;
	d[1] = v2->w * v1->vy - v1->w * v2->vy;
	d[2] = v2->w * v1->vz - v1->w * v2->vz;
	*dmax = sqrt( (d[0]*d[0]+d[1]*d[1]+d[2]*d[2])
		      / (v12[0]*v12[0]+v12[1]*v12[1]+v12[2]*v12[2]) );
      }
      else {
	*dmax = v1->w / v1->vn;	
      }
    }
    else {
      if ( uv31 >= 0 ) {
	*dmax = v2->w / v2->vn;	
      }
      else {
	fprintf( stdout, "%s: this case should never occur\n", proc );
      }
    }
  }
  
}



void _compute_3D_cone_min_max( Vector *v1, Vector *v2, Vector *v3, 
			       double *dmin, double *dmax )
{
  *dmin = v1->w / v1->vn;
  if ( *dmin > v2->w / v2->vn ) *dmin = v2->w / v2->vn;
  if ( *dmin > v3->w / v3->vn ) *dmin = v3->w / v3->vn;

  _compute_3D_cone_max( v1, v2, v3, dmax );
}



void _compute_3D_mask_min_max( VectorList *vl, ConeList *cl, 
			       double *dmin, double *dmax )
{
  int i;
  Vector *v1, *v2, *v3;
  double ldmax;
  
  *dmin = vl->allocatedVectors[0].w / vl->allocatedVectors[0].vn;
  
  for ( i=1; i<vl->n_vectors; i++ ) 
    if ( *dmin > vl->allocatedVectors[i].w / vl->allocatedVectors[i].vn )
      *dmin = vl->allocatedVectors[i].w / vl->allocatedVectors[i].vn;

  *dmax = *dmin;
  for ( i=0; i<cl->n_cones; i++ ) {
    v1 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv1 ]);
    v2 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv2 ]);
    v3 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv3 ]);

    _compute_3D_cone_max( v1, v2, v3, &ldmax );
    if ( *dmax < ldmax ) *dmax = ldmax;

  }
}


void _print_2D_cone( FILE *f, VectorList *vl, Cone *c )
{
  fprintf( f, "[" );
  _print_2D_vector( f, &(vl->allocatedVectors[c->iv1]) );
  fprintf( f, " " );
  _print_2D_vector( f, &(vl->allocatedVectors[c->iv2]) );
  fprintf( f, "]" );
}



void _print_3D_cone( FILE *f, VectorList *vl, Cone *c )
{
  fprintf( f, "[" );
  _print_3D_vector( f, &(vl->allocatedVectors[c->iv1]) );
  fprintf( f, " " );
  _print_3D_vector( f, &(vl->allocatedVectors[c->iv2]) );
  fprintf( f, " " );
  _print_3D_vector( f, &(vl->allocatedVectors[c->iv3]) );
  fprintf( f, "]" );
}













/*==================================================
 *
 * RESULTS
 *
 ==================================================*/



void _init_results_list( ResultList *rl )
{
  rl->error_optimal = 0.0;
  rl->n_results = 0;
  rl->n_allocated_results = 0;
  rl->allocatedResults = NULL;
}



void _free_results_list( ResultList *rl )
{ 
  int i;
  if ( rl->allocatedResults != NULL) {
    if ( rl->n_results > 0 ) {
      for ( i=0; i< rl->n_results; i++ )
	if ( rl->allocatedResults[i].w != NULL ) 
	  free( rl->allocatedResults[i].w );
    }
  }
  free( rl->allocatedResults );
  _init_results_list( rl );
}


void _sort_vectors_in_lexicographic_order( Vector **v, int left, int right )
{
  int i, last;
  Vector *t;
  int max, lmax;
  int sum, lsum;

  if ( left >= right ) return;

  t = v[left];   v[left] = v[(left+right)/2];   v[(left+right)/2] = t;
  
  last = left;

  lmax = v[left]->x;
  if ( lmax < v[left]->y) lmax = v[left]->y;
  if ( lmax < v[left]->z) lmax = v[left]->z;
  lsum = v[left]->x + v[left]->y + v[left]->z;

  for ( i = left+1; i <= right; i++ ) {
    max = v[i]->x;
    if ( max < v[i]->y) max = v[i]->y;
    if ( max < v[i]->z) max = v[i]->z;
    sum = v[i]->x + v[i]->y + v[i]->z;
    if ( (max < lmax) ||
	 (max == lmax && sum < lsum) ||
	 (max == lmax && sum == lsum
	  && ((v[i]->x > v[left]->x)
	      || (v[i]->x == v[left]->x && v[i]->y > v[left]->y)
	      || (v[i]->x == v[left]->x && v[i]->y == v[left]->y 
		  && v[i]->z > v[left]->z))) ) {
      t = v[++last];   v[last] = v[i];   v[i] = t;
    }
  }

  t = v[left];   v[left] = v[last];   v[last] = t;

  _sort_vectors_in_lexicographic_order( v, left, last-1 );
  _sort_vectors_in_lexicographic_order( v, last+1, right );
}


void _print_results_list( FILE *f, ResultList *rl,  VectorList *vl )
{
  char *proc = "_print_results_list";
  int *p;
  int i, j;
  int z;

  Vector **v = NULL;

  v = (Vector**)malloc( vl->n_vectors * sizeof(Vector*) );
  if ( v == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    return;
  }
  for ( i=0; i<vl->n_vectors; i++ ) {
    vl->allocatedVectors[i].index = i;
    v[i] = &(vl->allocatedVectors[i]);
  }
  _sort_vectors_in_lexicographic_order( v, 0, vl->n_vectors-1 );
  
  p = (int*)malloc( vl->n_vectors * sizeof(int) );
  if ( p == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    free( v );
    return;
  }
  for ( i=0; i<vl->n_vectors; i++ ) {
    p[i] = v[i]->index;
  }
  free( v );

  fprintf( f, "\n" );
  fprintf( f, " Best possible error = %f\n", rl->error_optimal );
  for ( i=0; i<vl->n_vectors; i++ ) 
    fprintf( f,"   [%2d]", vl->allocatedVectors[p[i]].x );
  fprintf( f, "\n" );
  for ( i=0; i<vl->n_vectors; i++ ) 
    fprintf( f,"   [%2d]", vl->allocatedVectors[p[i]].y );
  fprintf( f,"  ---       err      %%/opt      eps         time\n" );


  z = vl->allocatedVectors[0].z;
  for ( i=1; i<vl->n_vectors; i++ ) 
    if ( z < vl->allocatedVectors[p[i]].z ) z = vl->allocatedVectors[p[i]].z;
  if ( z > 0 ) {
    for ( i=0; i<vl->n_vectors; i++ ) 
      fprintf( f,"   [%2d]", vl->allocatedVectors[p[i]].z );
    fprintf( f, "\n" );
  }

 fprintf( f, "\n" );

 for ( j=0; j<rl->n_results; j++ ) {
   for ( i=0; i<vl->n_vectors; i++ ) 
     fprintf( f,"  %4d ", rl->allocatedResults[j].w[p[i]] );
   fprintf( f,"  ---    %9.6f (%6.1f)  %9.6f  %9.2f\n",
	    rl->allocatedResults[j].error * 100.0,
	    rl->allocatedResults[j].error * 100.0 / rl->error_optimal,
	    rl->allocatedResults[j].epsilon,
	    rl->allocatedResults[j].elapsed_time );
 }

 free( p );
}


void _print_results_list_in_latex( FILE *f, ResultList *rl,  VectorList *vl )
{
  char *proc = "_print_results_list_in_latex";
  int *p;
  int i, j;
  int z;
  int min, s;

  Vector **v = NULL;

  v = (Vector**)malloc( vl->n_vectors * sizeof(Vector*) );
  if ( v == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    return;
  }
  for ( i=0; i<vl->n_vectors; i++ ) {
    vl->allocatedVectors[i].index = i;
    v[i] = &(vl->allocatedVectors[i]);
  }
  _sort_vectors_in_lexicographic_order( v, 0, vl->n_vectors-1 );
  
  p = (int*)malloc( vl->n_vectors * sizeof(int) );
  if ( p == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    free( v );
    return;
  }
  for ( i=0; i<vl->n_vectors; i++ ) {
    p[i] = v[i]->index;
  }
  free( v );

  fprintf( f, "\n" );
  fprintf( f, "\\begin{table}\n" );
  fprintf( f, "\\begin{center}\n" );
  fprintf( f, "\\small\n" );
  fprintf( f, "\\begin{tabular}{|" );
  for ( i=0; i<vl->n_vectors; i++ ) {
    fprintf( f, "c" );
    if ( i<vl->n_vectors-1 ) fprintf( f, "|" );
  }
  fprintf( f, "|c|c|c|c|}\n" );
  fprintf( f, "\\hline\n" );
  
  z = vl->allocatedVectors[0].z;
  for ( i=1; i<vl->n_vectors; i++ ) 
    if ( z < vl->allocatedVectors[p[i]].z ) z = vl->allocatedVectors[p[i]].z;

  for ( i=0; i<vl->n_vectors; i++ ) {
    fprintf( f,"\\tiny " );
    fprintf( f,"$ " );
    fprintf( f,"\\!\\!\\!\\!" );
    if ( 0 ) fprintf( f,"\\left\( " );
    fprintf( f,"\\begin{array}{c}" );
    fprintf( f," %d \\\\ %d", vl->allocatedVectors[p[i]].x, vl->allocatedVectors[p[i]].y );
    if ( z > 0 ) 
      fprintf( f," \\\\ %d ", vl->allocatedVectors[p[i]].z );
    fprintf( f,"\\end{array} ");
    if ( 0 ) fprintf( f,"\\right) " );
    fprintf( f,"\\!\\!\\!\\!" );
    fprintf( f,"$ &\n" );
  }
  fprintf( f," error & \\%% / opt & $\\varepsilon$ & time \\\\ \\hline\n" );
  
 for ( j=0; j<rl->n_results; j++ ) {
   for ( i=0; i<vl->n_vectors; i++ ) {
     fprintf( f,"  %4d & ", rl->allocatedResults[j].w[p[i]] );
     if ( (i+1) % 10 == 0 ) fprintf( f,"\n" );
   }
   fprintf( f, "\n" );
   fprintf( f," %5.2f & %6.1f & %6.3f &\n",
	    rl->allocatedResults[j].error * 100.0,
	    rl->allocatedResults[j].error * 100.0 / rl->error_optimal,
	    rl->allocatedResults[j].epsilon);
   if ( rl->allocatedResults[j].elapsed_time < 1000 ) {
     fprintf( f, " %d ms ", (int)rl->allocatedResults[j].elapsed_time );
   }
   else {
     if ( rl->allocatedResults[j].elapsed_time < 60 * 1000 ) {
       fprintf( f, " %5.2f s ", rl->allocatedResults[j].elapsed_time / 1000 );
     }
     else {
       if ( rl->allocatedResults[j].elapsed_time < 10 * 60 * 1000 ) {
	 min = (int)(rl->allocatedResults[j].elapsed_time/(60*1000));
	 s = (int)((rl->allocatedResults[j].elapsed_time - min * 60 * 1000)/1000);
	 fprintf( f, " %d min %d s ", min, s );
       }
       else {
	 min = (int)(rl->allocatedResults[j].elapsed_time/(60*1000));
	 if ( min < 60 ) {
	   fprintf( f, " %d min ", min );
	 }
	 else {
	   fprintf( f, " %d h %d min ", min / 60, min % 60 );
	 }
       }
     }
   }
   fprintf( f," \\\\ \\hline\n" );
 }

  fprintf( f, "\\end{tabular}\n" );
  fprintf( f, "\\end{center}\n" );
  fprintf( f, "\\caption{Best possible error = %f.}\n", rl->error_optimal );
  fprintf( f, "\\end{table}\n\n" );


 free( p );
}


void _add_result_to_list(  ResultList *rl,  VectorList *vl,
			   double err, double dmin, double dmax, double eps,
			   double time, double init_time )
{
  char *proc = "_add_result_to_list";
  Result *newr;
  int nalloc;
  int i;

  if (  rl->n_results >= rl->n_allocated_results ) {

    if ( rl->n_allocated_results  <= 0 ) 
      nalloc = _alloc_;
    else
      nalloc = rl->n_allocated_results + _alloc_;

    newr = (Result*)malloc( nalloc * sizeof(Result) );
    if ( newr == NULL ) {
      fprintf( stderr, "%s: allocation error\n", proc );
      exit( -1 );
    }
    if ( rl->n_allocated_results  > 0 ) {
      memcpy( newr, rl->allocatedResults, 
	      rl->n_allocated_results * sizeof(Result) );
      free( rl->allocatedResults );
    }

    rl->allocatedResults = newr;
    rl->n_allocated_results = nalloc;
  }

  rl->allocatedResults[ rl->n_results ].w = (int*)malloc( vl->n_vectors * sizeof(int) );
  if ( rl->allocatedResults[ rl->n_results ].w == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    exit( -1 );
  }

  for ( i=0; i<vl->n_vectors; i++ )
    rl->allocatedResults[rl->n_results ].w[i] = vl->allocatedVectors[i].w;

  rl->allocatedResults[rl->n_results ].error = err;
  rl->allocatedResults[rl->n_results ].dmin  = dmin;
  rl->allocatedResults[rl->n_results ].dmax  = dmax;
  rl->allocatedResults[rl->n_results ].epsilon = eps;
  rl->allocatedResults[rl->n_results ].elapsed_time = 
    (time - init_time) * 10000 * 1000 / (double)CLOCKS_PER_SEC;

  rl->n_results ++;
}







/*==================================================
 *
 * VECTORS AND CONES
 *
 ==================================================*/



int _link_cones_to_vectors( VectorList *vl, ConeList *cl )
{
  int i, j;

  for ( i=0; i<vl->n_vectors; i++ ) 
    vl->allocatedVectors[i].n_cones = 0;

  for ( i=0; i<cl->n_cones; i++ ) {
    for ( j=0; j<vl->n_vectors; j++ ) {
      if ( j == cl->allocatedCones[i].iv1
	   || j == cl->allocatedCones[i].iv2
	   || j == cl->allocatedCones[i].iv3 )
	vl->allocatedVectors[j].n_cones ++;
    }
  }

  for ( i=0; i<vl->n_vectors; i++ ) {
    if ( vl->allocatedVectors[i].n_cones > 0 ) {
      vl->allocatedVectors[i].c = (Cone**)malloc( vl->allocatedVectors[i].n_cones *
						 sizeof( Cone* ) );
      if ( vl->allocatedVectors[i].c == NULL ) {
	fprintf( stderr, "Allocation error (#cones=%d)\n", 
		 vl->allocatedVectors[i].n_cones );
	exit( 0 );
      }
      vl->allocatedVectors[i].n_cones = 0;
    }
  }
  
  for ( i=0; i<cl->n_cones; i++ ) {
    for ( j=0; j<vl->n_vectors; j++ ) {
      if ( j == cl->allocatedCones[i].iv1
	   || j == cl->allocatedCones[i].iv2
	   || j == cl->allocatedCones[i].iv3 ) {
	vl->allocatedVectors[j].c[ vl->allocatedVectors[j].n_cones ] = &(cl->allocatedCones[i]);
	vl->allocatedVectors[j].n_cones ++;
      }
    }
  }

  return( 1 );
}



void _print_2D_vector_list( FILE *f, VectorList *vl )
{
  int i ,j;

  fprintf( f, "\n" );
  fprintf( f, "#VECTORS = %d", vl->n_vectors );
  fprintf( f, "\n" );
  for ( i=0; i<vl->n_vectors; i++ ) {
    fprintf( f, "v[#%3d] = ", i );
    _print_2D_vector( f, &(vl->allocatedVectors[i]) );
    fprintf( f, " c=%d", vl->allocatedVectors[i].n_cones );
    fprintf( f, " " );
    for ( j=0; j<vl->allocatedVectors[i].n_cones; j++ )
      _print_2D_cone( f, vl, vl->allocatedVectors[i].c[j] );
    fprintf( f, "\n" );
  }
  fprintf( f, "\n" );
}



void _print_3D_vector_list( FILE *f, VectorList *vl )
{
  int i ,j;

  fprintf( f, "\n" );
  fprintf( f, "#VECTORS = %d", vl->n_vectors );
  fprintf( f, "\n" );
  for ( i=0; i<vl->n_vectors; i++ ) {
    fprintf( f, "v[#%3d] = ", i );
    _print_3D_vector( f, &(vl->allocatedVectors[i]) );
    fprintf( f, " c=%d", vl->allocatedVectors[i].n_cones );
    fprintf( f, " " );
    for ( j=0; j<vl->allocatedVectors[i].n_cones; j++ ) {
      _print_3D_cone( f, vl, vl->allocatedVectors[i].c[j] );
      if ( j<vl->allocatedVectors[i].n_cones-1 && j>0 && (j+1)%2 == 0 )
	fprintf( f, "\n                        ");
    }
    fprintf( f, "\n" );
  }
  fprintf( f, "\n" );
}



void _print_cone_list( FILE *f, VectorList *vl, ConeList *cl )
{
  int i;
  double emin, emax;
  
  emin = emax= cl->allocatedCones[0].error;
  for ( i=1; i<cl->n_cones; i++ ) {
    if ( emin > cl->allocatedCones[i].error )
      emin = cl->allocatedCones[i].error;
    if ( emax < cl->allocatedCones[i].error )
      emax = cl->allocatedCones[i].error;
  }

   
  fprintf( f, "\n" );
  fprintf( f, "#CONES = %d", cl->n_cones );
  fprintf( f, "   ERR MIN = %f", emin );
  fprintf( f, "   ERR MAX = %f", emax );
  fprintf( f, "\n" );
  for ( i=0; i<cl->n_cones; i++ ) {
    fprintf( f, "c[#%3d] = ", i );
    if ( cl->allocatedCones[i].iv3 < 0 )
      _print_2D_cone( f, vl, &(cl->allocatedCones[i]) );
    else {
      if ( _verbose_ >= 2 ) {
	switch ( cl->allocatedCones[i].type ) {
	case _UNSET_  : fprintf( f, "--- " ); break;
	case _INSIDE_ : fprintf( f, "IJK " ); break;
	case _OUT_IJ_ : fprintf( f, "IJ- " ); break;
	case _OUT_JK_ : fprintf( f, "-JK " ); break;
	case _OUT_KI_ : fprintf( f, "I-K " ); break;
	}
    }
      _print_3D_cone( f, vl, &(cl->allocatedCones[i]) );
    }
    fprintf( f, ", E=%f", cl->allocatedCones[i].error );
    if ( _verbose_ >= 2 )
      fprintf( f, ", H=%f", (1-cl->allocatedCones[i].error)/(1+cl->allocatedCones[i].error) );
    fprintf( f, "\n" );	
  }
  fprintf( f, "\n" );
}



void _print_cone_information( FILE *fout, double optimal_err, 
			       double previous_err,
			       double err, double dmin,
			       double dmax, double epsilon,
			      double ct, double it )
{
  fprintf( fout, "\n" );
  fprintf( fout, "Error = %f   %g %% / OPT --- Epsilon = %f", 
	   err, 100.0*err/optimal_err, epsilon );
  fprintf( fout, " --- Elapsed time = %f ms",
	   (ct - it) * 10000 * 1000 / (double)CLOCKS_PER_SEC );
  fprintf( fout, "\n" );
  if ( _verbose_ >= 1 ) {
    fprintf( fout, "--- Incr = %g\n", previous_err - err );
  }
  fprintf( fout, "--- dmin = %f   dmax = %f", dmin, dmax );
  fprintf( fout, "\n" );
}



void _print_mask_information( FILE *fout, double optimal_err, 
			       double previous_err,
			       double err, double dmin,
			       double dmax, double epsilon,
			       double lerr, double ldmin,
			       double ldmax, double lepsilon,
			      double ct, double it )
{
  fprintf( fout, "\n" );
  fprintf( fout, "Error = %f   %g %% / OPT --- Epsilon = %f", 
	   lerr, 100.0*lerr/optimal_err, lepsilon );
  fprintf( fout, " --- Elapsed time = %f ms",
	   (ct - it) * 10000 * 1000 / (double)CLOCKS_PER_SEC );
  fprintf( fout, "\n" );
  if ( _verbose_ >= 1 ) {
    fprintf( fout, "--- Incr = %g\n", previous_err - lerr );
    fprintf( fout, "--- Expected = %f   Diff = %g\n", 
	     err, lerr-err);
  }
  fprintf( fout, "--- dmin = %f   dmax = %f", ldmin, ldmax );
  if ( _verbose_ >= 1 ) {
    fprintf( fout, "   (Exp dmin = %f dmax =%f)", dmin, dmax );
  }
  fprintf( fout, "\n" );
}



void _print_2D_weighted_vector_list( FILE *f, VectorList *vl, int n )
{
  int i;
  for (i=0; i<n; i++ ) {
    fprintf( f, "    v[#%2d] = ", i );
    _print_2D_vector( f, &(vl->allocatedVectors[i]) );
    fprintf( f, "   w=%4d\n", vl->allocatedVectors[i].w );
  }
}



void _print_3D_weighted_vector_list( FILE *f, VectorList *vl, int n )
{
  int i;
  for (i=0; i<n; i++ ) {
    fprintf( f, "    v[#%2d] = ", i );
    _print_3D_vector( f, &(vl->allocatedVectors[i]) );
    fprintf( f, "   w=%4d\n", vl->allocatedVectors[i].w );
  }
}











/*==================================================
 *
 * 2D MASK
 *
 ==================================================*/



void _init_2D_mask( VectorList *vl, ConeList *cl, double *s ) 
{
  int i1, i2;
  
  i1 = _add_vector_to_list( vl, 1, 0, 0, s );
  i2 = _add_vector_to_list( vl, 1, 1, 0, s );

  (void)_add_cone_to_list( vl, cl, i1, i2, -1 );
}



void _build_2D_mask( VectorList *vl, ConeList *cl, double *s, int size ) 
{
  int d = (size-1)/2;
  int ic, iv, i, k;
  int nx, ny;
  Cone *c;

  /* loop on depth
   */
  for ( k=2; k<=d; k++ ) {

    /* loop on cones
     */
    for ( i=0; i<cl->n_cones; i++ ) {

      /* do not split
       */
      c = &(cl->allocatedCones[i]);
      nx = vl->allocatedVectors[c->iv1].x + vl->allocatedVectors[c->iv2].x;
      ny = vl->allocatedVectors[c->iv1].y + vl->allocatedVectors[c->iv2].y;
      if ( nx > k || ny > k ) continue;

      /* create new vector
       */
      iv = _add_vector_to_list( vl, nx, ny, 0, s );
      
      /* create new cone
       */
      ic =_add_cone_to_list( vl, cl, iv, c->iv2, -1 );
      _move_cone_backward( cl, ic );
      
      /* change the considered cone
       */
      cl->allocatedCones[i].iv2 = iv;
      _compute_2D_cone_error( &(cl->allocatedCones[i]),
			      &(vl->allocatedVectors[cl->allocatedCones[i].iv1]),
			      &(vl->allocatedVectors[cl->allocatedCones[i].iv2]) );
      _move_cone_forward( cl, i );
      
      /* consider it again
       */
      i--;
    }

  }

}



void _build_2D_mask_wrt_error( FILE *fout, 
			       VectorList *vl, ConeList *cl, 
			       double *s, int size, int n_vectors, 
			       double error ) 
{
  int d = (size-1)/2;
  int ic, iv, i=0;
  int nx, ny;
  Cone *c;

  if ( _trace_ ) {
    _print_2D_vector_list( fout, vl );
    _print_cone_list( fout, vl, cl );
  }

  /* loop on cones
   */
  while ( i<cl->n_cones && vl->n_vectors< n_vectors 
	  && cl->allocatedCones[i].error > error ) {

    /* do not split
     */
    c = &(cl->allocatedCones[i]);
    nx = vl->allocatedVectors[c->iv1].x + vl->allocatedVectors[c->iv2].x;
    ny = vl->allocatedVectors[c->iv1].y + vl->allocatedVectors[c->iv2].y;
    if ( nx > d || ny > d ) {
      i++;
      continue;
    }
    
    /* create new vector
     */
    iv = _add_vector_to_list( vl, nx, ny, 0, s );
      
    /* create new cone
     */
    ic =_add_cone_to_list( vl, cl, iv, c->iv2, -1 );
    _move_cone_backward( cl, ic );
      
    /* change the considered cone
     */
    cl->allocatedCones[i].iv2 = iv;
    _compute_2D_cone_error( &(cl->allocatedCones[i]),
			    &(vl->allocatedVectors[cl->allocatedCones[i].iv1]),
			    &(vl->allocatedVectors[cl->allocatedCones[i].iv2]) );
    _move_cone_forward( cl, i );
  
    if ( _trace_ ) {
      fprintf( fout, "add vector (%2d %2d)", nx, ny );
      fprintf( fout, " => maximal error becomes %f\n", 
	       cl->allocatedCones[0].error );
    }
    
  }

}




void _symmetrise_2D_mask( VectorList *vl, ConeList *cl, double *s )
{
  int i, nc;
  int iv1, iv2;
  Cone *c;

  
  /* add cones
   */
  nc = cl->n_cones;
  for ( i=0; i<nc;  i++ ) {

    c = &(cl->allocatedCones[ i ]);

    iv1 = _add_vector_to_list( vl, vl->allocatedVectors[c->iv1].y, 
			       vl->allocatedVectors[c->iv1].x, 0, s );
    iv2 = _add_vector_to_list( vl, vl->allocatedVectors[c->iv2].y, 
			       vl->allocatedVectors[c->iv2].x, 0, s );

    /* change vector order to ensure
       positivity of the vector product
    */
    (void)_add_cone_to_list( vl, cl, iv2, iv1, -1 );
    
  }

}



void _sort_2D_vectors( Vector **v, VectorList *vl, ConeList *cl )
{
  int i, j, n;
  
  for ( i=0; i<vl->n_vectors; i++ ) 
    vl->allocatedVectors[i].flag = 0;
  for ( i=0; i<cl->n_cones; i++ ) 
    cl->allocatedCones[i].flag = 0;
  
  /* first cone
   */

  if ( _norme2v_2D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
       < _norme2v_2D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) ) ) {
    v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
    v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
  }
  else {
    v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
    v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
  }

  cl->allocatedCones[0].flag = 1;
  v[0]->flag = 1;
  v[1]->flag = 1;

  n = 2;

  for ( ; n < vl->n_vectors;  ) {

    for ( j=0, i=1; i<cl->n_cones && j==0; i++ ) {
      if ( cl->allocatedCones[i].flag == 1 ) continue;
      if ( vl->allocatedVectors[cl->allocatedCones[i].iv1].flag == 0 
	   && vl->allocatedVectors[cl->allocatedCones[i].iv2].flag == 0 )
	continue;
      j = i;
    }

    if ( j > 0 ) {
      cl->allocatedCones[j].flag = 1;
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv1].flag == 0 ) {
	v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv1]);
	v[n]->flag = 1;
	n++;
      }
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv2].flag == 0 ) {
	v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv2]);
	v[n]->flag = 1;
	n++;
      }
    }      

  }
}



void _compute_2D_weights( FILE *fout, ResultList *rl, 
			  VectorList *vl, ConeList *cl,
			  int f, int l ) 
{
  char *proc = "_compute_2D_weights";
  Vector **v;
  int i, j;
  double dmin, dmax, eps;
  double perr=1.0;
  double erropt, err;
  int iv;

  double v1min, v1max;

  int ok;
  Cone * cone;
  double ldmin, ldmax, lerr, leps;

  struct tms timestruct;
  double it, ct;



  /* add links from vectors to cones
   */
  (void)_link_cones_to_vectors( vl, cl );

  /* allocate and sort the vectors
   */
  v = (Vector**)malloc( vl->n_vectors * sizeof(Vector*) );
  if ( v == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    return;
  }
  _sort_2D_vectors( v, vl, cl );



  /* initialization
     flag = 0 -> no set yet
     flag = 1 -> already set
  */

  for ( i=0; i<vl->n_vectors; i++ ) 
    vl->allocatedVectors[i].flag = 0;
  for ( i=0; i<cl->n_cones; i++ ) 
    cl->allocatedCones[i].flag = 0;


  /* the largest error
   */
  erropt = cl->allocatedCones[0].error;


  /* set bounds for the vectors of the first cone
     low bound = f
     high bound = l * |v|_1
   */
  v[0]->f = floor(f * _normeIv_2D( v[0] ) );
  v[0]->l = (int)(l * _norme1v_2D( v[0] ) );

  v[1]->f = floor(f * _normeIv_2D( v[1] ) );
  v[1]->l = (int)(l * _norme1v_2D( v[1] ) );

  v[0]->flag = 1;
  v[1]->flag = 1;



  it = times( &timestruct );



  /* loops for the first cone
   */
  for ( v[0]->w  = v[0]->f;
	v[0]->w <= v[0]->l;
	v[0]->w ++ ) {
    if ( _trace_ )
      fprintf( fout,"%3d   [%d %d]\n", v[0]->w, v[0]->f, v[0]->l );

    if ( _criterium_ == _MAXIMALITY_ ) {
      /* get better bounds for the second vector
	 to satisfy the maximality conditions
      */
      v1min = v[0]->w 
	* (v[0]->vx * v[1]->vx + v[0]->vy * v[1]->vy)
	/ (v[0]->vx * v[0]->vx + v[0]->vy * v[0]->vy);
      v1max = v[0]->w
	* (v[1]->vx * v[1]->vx + v[1]->vy * v[1]->vy)
	/ (v[0]->vx * v[1]->vx + v[0]->vy * v[1]->vy);
      
      v[1]->f = ceil( v1min );
      v[1]->l = floor( v1max );
    }
    
    for ( v[1]->w  = v[1]->f;
	  v[1]->w <= v[1]->l;
	  v[1]->w ++ ) {
      if ( _trace_ )
	fprintf( fout, "%3d %3d   [%d %d]", v[0]->w, v[1]->w, v[1]->f, v[1]->l );

      /* check condition
	 => no more necessary for convexity, 
	 since bounds on v[1] are computed so that it is satisfied
       */
      switch( _criterium_ ) {
      case _MAXIMALITY_ :
	/*
	  if ( _check_2D_maximality( vl, &(cl->allocatedCones[0]) ) == 0 ) {
	  continue; 
	  }
	*/
	break;
      case _CONVEXITY_ :
	if ( _check_2D_convexity( vl, cl, &(cl->allocatedCones[0]) ) == 0 ) {
	  if ( _trace_ )
	    fprintf( fout, " cone #1 not convex\n" );
	  continue; 
	}
	break;   
      }
      if ( _trace_ )
	fprintf( fout, "\n" );



      /* compute min and max distance, epsilon and error
      */      
      _compute_2D_cone_min_max( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]),
				&(vl->allocatedVectors[cl->allocatedCones[0].iv2]),
				&dmin, &dmax );
      eps = (dmin + dmax)/2.0;
      err = dmax/eps - 1;


      
      /* if we do not perform better,
	 go to next
      */
      if ( err >= perr ) {
	if ( _trace_ ) {
	  fprintf( fout,"%3d %3d too large error", v[0]->w, v[1]->w );
	  fprintf( fout,"\n" );
	}
	continue;
      }

      /* Here, we've got better candidates
	 for the cone of largest error
      */

      /* special case:
	 there is only one cone
      */
      iv = 2;
      if ( iv == vl->n_vectors ) {
	ct = times( &timestruct );
	_add_result_to_list( rl, vl,  err, dmin, dmax, eps, ct, it );
	_print_cone_information( fout, erropt, perr, err, dmin, dmax, eps, ct, it );
	_print_2D_weighted_vector_list( fout, vl, 2 );
	perr = err;
	continue;
      }

      cl->allocatedCones[0].flag = 1;

      /* here we have more work to do
	 compute the bounds for the other vectors
	 from the error of the first cone
       */
      for ( i=2, ok=1; i<vl->n_vectors && ok==1; i++ ) {

	if ( _be_strict_ == 0 ) {
	  v[i]->f = (int)(floor( eps*(1-perr)*v[i]->vn ) + 0.5 );
	  v[i]->l = (int)(ceil(  eps*(1+perr)*v[i]->vn ) + 0.5 );
	}
	else {
	  v[i]->f = (int)(ceil(  eps*(1-perr)*v[i]->vn ) + 0.5 );
	  v[i]->l = (int)(floor( eps*(1+perr)*v[i]->vn ) + 0.5 );
	}

	if ( v[i]->f <= 0 ) v[i]->f = 1;
	if ( v[i]->l > (int)(l * _norme1v_2D(v[i])) ) 
	  v[i]->l = (int)(l * _norme1v_2D(v[i]));
	
	v[i]->w = v[i]->f-1;

	if ( _verbose_ >= 3 ) {
	  fprintf( stdout, "  bounds #%2d: %3d -> %3d\n", i,
		   v[i]->f, v[i]->l );
	}
	if ( v[i]->l < v[i]->f ) ok = 0;
      }

      if ( ok == 0 ) {
	if ( _trace_ ) 
	  fprintf( fout,"%3d %3d empty intervals\n", v[0]->w, v[1]->w );
	if ( _verbose_ >= 3 ) {
	  fprintf( stdout, "  empty intervals\n" );
	}
	continue;
      }


      /* bounds have been computed
	 we have to find a weight set that is optimal
       */
      do {
	
	/* find a weight that satisfy conditions
	   for next vector
	*/
	do {
	  v[iv]->w ++;
	  if ( _trace_ ) {
	    if ( v[iv]->w == v[iv]->f ) {
	      fprintf( fout,"%3d %3d", v[0]->w, v[1]->w );
	      for (i=2; i<=iv; i++ )
		fprintf( fout," %3d", v[i]->w );
	      fprintf( fout,"   [%d %d]\n", v[iv]->f, v[iv]->l );
	    }
	  }
	  if ( v[iv]->w <= v[iv]->l ) {
	    v[iv]->flag = 1;
	    ok = 1;
	    /* look at the cones that share the vector
	       we work with
	    */
	    for ( j = 0; j<v[iv]->n_cones && ok == 1; j++ ) {
	      cone = v[iv]->c[j];
	      if ( cone->flag == 1 ) continue;
	      if ( (&(vl->allocatedVectors[cone->iv1]) == v[iv] 
		    && vl->allocatedVectors[cone->iv2].flag == 1)
		   || (vl->allocatedVectors[cone->iv1].flag == 1 
		       && &(vl->allocatedVectors[cone->iv2]) == v[iv] ) ) {
		switch( _criterium_ ) {
		case _MAXIMALITY_ :
		  ok = ok &&  _check_2D_maximality( vl, cone );
		  break;
		case _CONVEXITY_ :
		  ok = ok && _check_2D_convexity( vl, cl, cone );
		  break;   
		}
		if ( _trace_ && ok == 0 ) {
		    fprintf( fout,"%3d %3d", v[0]->w, v[1]->w );
		    for (i=2; i<=iv; i++ )
		      fprintf( fout," %3d", v[i]->w );
		    fprintf( fout,"   not maximal/convex\n" );
		}
		if ( ok == 1 ) cone->flag = 1;
	      }
	    }
	  }
	} while ( v[iv]->w <= v[iv]->l && ok == 0 );

	/* if not possible
	   get back to previous vector
	*/
	if ( v[iv]->w > v[iv]->l ) {
	  if ( _verbose_ >= 3 ) {
	    fprintf( stdout, "   TEST FAILURE " );
	    for (j=2; j<iv; j++ ) {
	      fprintf( stdout, "#%d=%d ", j, v[j]->w );
	    }
	    fprintf( stdout, "#%d=\n", iv );
	  }
	  v[iv]->w = v[iv]->f-1;
	  for ( j = 0; j<v[iv]->n_cones; j++ )
	    v[iv]->c[j]->flag = 0;
	  v[iv]->flag = 0;
	  iv --;
	}
	/* or get back to next
	 */
	else {
	  if ( iv == vl->n_vectors-1 ) {
	    /* we got a mask
	     */
	    if ( _verbose_ >= 3 ) {
	      fprintf( stdout, "   TEST SUCCESS " );
	      for (j=2; j<=iv; j++ ) {
		fprintf( stdout, "#%d=%d ", j, v[j]->w );
	      }
	      fprintf( stdout, "\n" );
	    }
	    _compute_2D_mask_min_max( vl, cl, &ldmin, &ldmax );
	    leps = (ldmin + ldmax)/2.0;
	    lerr = ldmax/leps - 1;
	    
	    if ( 1.0001 * lerr < perr ) {
	      ct = times( &timestruct );
	      _add_result_to_list( rl, vl,  lerr, ldmin, ldmax, leps, ct, it );
	      _print_mask_information( fout, erropt, perr, 
					err, dmin, dmax, eps,
					lerr, ldmin, ldmax, leps, ct, it );
	      _print_2D_weighted_vector_list( fout, vl, vl->n_vectors );
	      perr = lerr;
	    }
	  }
	  else {
	    v[iv]->flag = 1;
	    iv ++;
	  }
	}

      } while ( iv > 1 );

      
      /* perr = err; */

    } /* for v[1] */
  }   /* for v[0] */

  free( v );
}













/*==================================================
 *
 * 3D MASK
 *
 ==================================================*/



void _init_3D_mask( VectorList *vl, ConeList *cl, double *s ) 
{
  int i1, i2, i3;

  i1 = _add_vector_to_list( vl, 1, 0, 0, s );
  i2 = _add_vector_to_list( vl, 1, 1, 0, s );
  i3 = _add_vector_to_list( vl, 1, 1, 1, s );

  (void)_add_cone_to_list( vl, cl, i1, i2, i3 );
}





void _symmetrise_3D_mask_wrt_vector( VectorList *vl, ConeList *cl, double *s,
				     Vector *normal, int first, int last )
{
  int i;
  int iv1, iv2, iv3;
  Cone *c;
  Vector *v;
  int ps, x, y, z;

  int n;

  n = normal->x * normal->x + normal->y * normal->y + normal->z * normal->z;

  for ( i=first; i<=last;  i++ ) {

    c = &(cl->allocatedCones[ i ]);

    /* vector #1 */
    v = &(vl->allocatedVectors[c->iv1]);
    ps = normal->x * v->x + normal->y * v->y + normal->z * v->z;
    x = v->x - 2 * ps * normal->x / n;
    y = v->y - 2 * ps * normal->y / n;
    z = v->z - 2 * ps * normal->z / n;
    iv1 = _add_vector_to_list( vl, x, y, z, s );

    /* vector #2 */
    v = &(vl->allocatedVectors[c->iv2]);
    ps = normal->x * v->x + normal->y * v->y + normal->z * v->z;
    x = v->x - 2 * ps * normal->x / n;
    y = v->y - 2 * ps * normal->y / n;
    z = v->z - 2 * ps * normal->z / n;
    iv2 = _add_vector_to_list( vl, x, y, z, s );

    /* vector #3 */
    v = &(vl->allocatedVectors[c->iv3]);
    ps = normal->x * v->x + normal->y * v->y + normal->z * v->z;
    x = v->x - 2 * ps * normal->x / n;
    y = v->y - 2 * ps * normal->y / n;
    z = v->z - 2 * ps * normal->z / n;
    iv3 = _add_vector_to_list( vl, x, y, z, s );


    /* change vector order to ensure
       positivity of the vector product
    */
    (void)_add_cone_to_list( vl, cl, iv2, iv1, iv3 );
    
  }

}

void _symmetrise_3D_mask( VectorList *vl, ConeList *cl, double *voxel_size,
			  int anisotropy_y, int anisotropy_z )
{
  int nc1, nc0 = cl->n_cones;
  Vector normal;

  if ( anisotropy_z ) {
    /* add cones wrt plane (1,0,0) (0,1,1)
       the vectorial product of the two above vectors 
       is (0,-1,1)
    */
    normal.x =  0;
    normal.y = -1;
    normal.z =  1;
    _symmetrise_3D_mask_wrt_vector( vl, cl, voxel_size, 
				    &normal, 0, nc0-1 );
    nc1 = cl->n_cones;
    
    /* add cones wrt plane (0,1,0) (1,0,1)
       the vectorial product of the two above vectors 
       is (1,0,-1)
    */
    normal.x =  1;
    normal.y =  0;
    normal.z = -1;
    _symmetrise_3D_mask_wrt_vector( vl, cl, voxel_size, 
				    &normal, nc0, nc1-1 );
    if ( anisotropy_y ) {
      /* add cones wrt plane (1,1,0) (0,0,1)
	 the vectorial product of the two above vectors 
	 is (1,-1,0)
      */
      normal.x =  1;
      normal.y = -1;
      normal.z =  0;
      _symmetrise_3D_mask_wrt_vector( vl, cl, voxel_size, 
				      &normal, 0, cl->n_cones-1 );
    }
  }
  else {
    if ( anisotropy_y ) {
      /* add cones wrt plane (1,1,0) (0,0,1)
	 the vectorial product of the two above vectors 
	 is (1,-1,0)
      */
      normal.x =  1;
      normal.y = -1;
      normal.z =  0;
      _symmetrise_3D_mask_wrt_vector( vl, cl, voxel_size, 
				      &normal, 0, nc0-1 );
      nc1 = cl->n_cones;
      
      /* add cones wrt plane (0,1,0) (1,0,1)
	 the vectorial product of the two above vectors 
	 is (1,0,-1)
      */
      normal.x =  1;
      normal.y =  0;
      normal.z = -1;
      _symmetrise_3D_mask_wrt_vector( vl, cl, voxel_size, 
				      &normal, nc0, nc1-1 );
    }
  }
}



void _split_3D_cone_along_JK( VectorList *vl, ConeList *cl, double *s, int i )
{
  Cone *c;
  int iv, ic;
  
  /* get or create new vector 
   */
  c = &(cl->allocatedCones[i]);
  iv = _add_vector_to_list( vl, 
			    vl->allocatedVectors[c->iv2].x + vl->allocatedVectors[c->iv3].x,
			    vl->allocatedVectors[c->iv2].y + vl->allocatedVectors[c->iv3].y, 
			    vl->allocatedVectors[c->iv2].z + vl->allocatedVectors[c->iv3].z, s );
 
  
  /* create new cone 
   */
  ic = _add_cone_to_list( vl, cl, c->iv1, iv, c->iv3 );
  _move_cone_backward( cl, ic );
  
  /* change the considered cone */
  cl->allocatedCones[i].iv3 = iv;
  _compute_3D_cone_error( &(cl->allocatedCones[i]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv1]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv2]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv3]) );
  _move_cone_forward( cl, i );
}



void _split_3D_cone_along_KI( VectorList *vl, ConeList *cl, double *s, int i )
{
  Cone *c;
  int iv, ic;

  /* get or create new vector 
   */
  c = &(cl->allocatedCones[i]);
  iv = _add_vector_to_list( vl, 
			    vl->allocatedVectors[c->iv3].x + vl->allocatedVectors[c->iv1].x,
			    vl->allocatedVectors[c->iv3].y + vl->allocatedVectors[c->iv1].y, 
			    vl->allocatedVectors[c->iv3].z + vl->allocatedVectors[c->iv1].z, s );
  
  /* create new cone 
   */
  ic = _add_cone_to_list( vl, cl, c->iv1, c->iv2, iv );
  _move_cone_backward( cl, ic );
	
  /* change the considered cone */
  cl->allocatedCones[i].iv1 = iv;
  _compute_3D_cone_error( &(cl->allocatedCones[i]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv1]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv2]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv3]) );
  _move_cone_forward( cl, i );
}



void _split_3D_cone_along_IJ( VectorList *vl, ConeList *cl, double *s, int i )
{
  Cone *c;
  int iv, ic;

  /* get or create new vector 
   */
  c = &(cl->allocatedCones[i]);
  iv = _add_vector_to_list( vl, 
			    vl->allocatedVectors[c->iv1].x + vl->allocatedVectors[c->iv2].x,
			    vl->allocatedVectors[c->iv1].y + vl->allocatedVectors[c->iv2].y, 
			    vl->allocatedVectors[c->iv1].z + vl->allocatedVectors[c->iv2].z, s );
  
  /* create new cone 
   */
  ic = _add_cone_to_list( vl, cl, iv, c->iv2, c->iv3 );
  _move_cone_backward( cl, ic );

  /* change the considered cone */
  cl->allocatedCones[i].iv2 = iv;
  _compute_3D_cone_error( &(cl->allocatedCones[i]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv1]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv2]),
			  &(vl->allocatedVectors[cl->allocatedCones[i].iv3]) );
  _move_cone_forward( cl, i );
}

void _build_3D_mask_XxXx5( VectorList *vl, ConeList *cl, double *s, int size ) 
{
  Cone *cc;
  int iv1, iv2, iv3;
  int adjcone, adjside;
  
  cc = &(cl->allocatedCones[0]);
  iv1 = cc->iv3;
  iv2 = cc->iv1;
  iv3 = cc->iv2;	
  _split_3D_cone_along_KI( vl, cl, s, 0 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  if ( size == 335 ) return;

  cc = &(cl->allocatedCones[0]);
  iv1 = cc->iv1;
  iv2 = cc->iv2;
  iv3 = cc->iv3;
  _split_3D_cone_along_IJ( vl, cl, s, 0 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  if ( size == 355 ) return;

  cc = &(cl->allocatedCones[1]);
  iv1 = cc->iv2;
  iv2 = cc->iv3;
  iv3 = cc->iv1;
  _split_3D_cone_along_JK( vl, cl, s, 1 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

}

void _build_3D_mask_3x3x5( VectorList *vl, ConeList *cl, double *s ) 
{
  _build_3D_mask_XxXx5( vl, cl, s, 335 );
}

void _build_3D_mask_3x5x5( VectorList *vl, ConeList *cl, double *s ) 
{
  _build_3D_mask_XxXx5( vl, cl, s, 355 );
}

void _build_3D_mask_5x5x5( VectorList *vl, ConeList *cl, double *s ) 
{
  _build_3D_mask_XxXx5( vl, cl, s, 555 );
}



void _build_3D_mask_7x7x7( VectorList *vl, ConeList *cl, double *s ) 
{
  Cone *cc;
  int iv1, iv2, iv3;
  int adjcone, adjside;

  _build_3D_mask_5x5x5( vl, cl, s );

  cc = &(cl->allocatedCones[0]);
  iv1 = cc->iv3;
  iv2 = cc->iv1;
  iv3 = cc->iv2;
  _split_3D_cone_along_KI( vl, cl, s, 0 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[2]);
  iv1 = cc->iv1;
  iv2 = cc->iv2;
  iv3 = cc->iv3;
  _split_3D_cone_along_IJ( vl, cl, s, 2 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[4]);
  iv1 = cc->iv2;
  iv2 = cc->iv3;
  iv3 = cc->iv1;
  _split_3D_cone_along_JK( vl, cl, s, 4 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[4]);
  iv1 = cc->iv3;
  iv2 = cc->iv1;
  iv3 = cc->iv2;	
  _split_3D_cone_along_KI( vl, cl, s, 4 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[1]);
  iv1 = cc->iv2;
  iv2 = cc->iv3;
  iv3 = cc->iv1;
  _split_3D_cone_along_JK( vl, cl, s, 1 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[0]);
  iv1 = cc->iv2;
  iv2 = cc->iv3;
  iv3 = cc->iv1;
  _split_3D_cone_along_JK( vl, cl, s, 0 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }

  cc = &(cl->allocatedCones[3]);
  iv1 = cc->iv1;
  iv2 = cc->iv2;
  iv3 = cc->iv3;
  _split_3D_cone_along_IJ( vl, cl, s, 3 );
  adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
  switch( adjside ) {
  default : break;
  case 0 :
    _split_3D_cone_along_JK( vl, cl, s, adjcone );
    break;
  case 1 :
    _split_3D_cone_along_KI( vl, cl, s, adjcone );
    break;
  case 2 :
    _split_3D_cone_along_IJ( vl, cl, s, adjcone );
    break;
  }


}


  
void _build_3D_mask( VectorList *vl, ConeList *cl, double *s, int size ) 
{
  char *proc = "_build_3D_mask";
  int d = (size-1)/2;
  int ok[3];
  double error[3], minerror = 0;
  int split;
  Vector sum;
  Cone *cc, csum;
  int i, k;

  int iv1, iv2, iv3;
  int adjcone, adjside;


  /* loop on depth
   */
  for ( k=2; k<=d; k++ ) {
    
    /* loop on cones
     */
    for ( i=0; i<cl->n_cones; i++ ) {

      cc = &(cl->allocatedCones[i]);

      ok[0] = ok[1] = ok[2] = 1;

      /* is it possible to split ?
       */
      if ( vl->allocatedVectors[cc->iv1].x + vl->allocatedVectors[cc->iv2].x > k 
	   || vl->allocatedVectors[cc->iv1].y + vl->allocatedVectors[cc->iv2].y > k 
	   || vl->allocatedVectors[cc->iv1].z + vl->allocatedVectors[cc->iv2].z > k  )
	ok[2] = 0;
      if ( vl->allocatedVectors[cc->iv2].x + vl->allocatedVectors[cc->iv3].x > k 
	   || vl->allocatedVectors[cc->iv2].y + vl->allocatedVectors[cc->iv3].y > k 
	   || vl->allocatedVectors[cc->iv2].z + vl->allocatedVectors[cc->iv3].z > k  )
	ok[0] = 0;
      if ( vl->allocatedVectors[cc->iv3].x + vl->allocatedVectors[cc->iv1].x > k 
	   || vl->allocatedVectors[cc->iv3].y + vl->allocatedVectors[cc->iv1].y > k 
	   || vl->allocatedVectors[cc->iv3].z + vl->allocatedVectors[cc->iv1].z > k  )
	ok[1] = 0;
	
      if ( ok[0] == 0 && ok[1] == 0 && ok[2] == 0 ) 
	continue;

      /* check assumptions
       */
      split = -1;
      if ( ok[0] == 1 ) {

	_init_vector( &sum );
	sum.x = vl->allocatedVectors[cc->iv2].x + vl->allocatedVectors[cc->iv3].x;
	sum.y = vl->allocatedVectors[cc->iv2].y + vl->allocatedVectors[cc->iv3].y;
	sum.z = vl->allocatedVectors[cc->iv2].z + vl->allocatedVectors[cc->iv3].z;
	_compute_3D_uv_vectors( &sum, s );

	_init_cone( &csum );
	_compute_3D_cone_error( &csum, 
				&(vl->allocatedVectors[cc->iv1]), 
				&sum,
				&(vl->allocatedVectors[cc->iv3]) );
	error[0] = csum.error;

	_compute_3D_cone_error( &csum, 
				&(vl->allocatedVectors[cc->iv1]), 
				&(vl->allocatedVectors[cc->iv2]), 
				&sum );
	if ( error[0] < csum.error ) error[0] = csum.error;
	
	split = 0;
	minerror = error[0];

      }

      if ( ok[1] == 1 ) {

	_init_vector( &sum );
	sum.x = vl->allocatedVectors[cc->iv3].x + vl->allocatedVectors[cc->iv1].x;
	sum.y = vl->allocatedVectors[cc->iv3].y + vl->allocatedVectors[cc->iv1].y;
	sum.z = vl->allocatedVectors[cc->iv3].z + vl->allocatedVectors[cc->iv1].z;
	_compute_3D_uv_vectors( &sum, s );

	_init_cone( &csum );
	_compute_3D_cone_error( &csum, 
				&(vl->allocatedVectors[cc->iv1]), 
				&(vl->allocatedVectors[cc->iv2]), 
				&sum );
	error[1] = csum.error;

	_compute_3D_cone_error( &csum, 
				&sum,
				&(vl->allocatedVectors[cc->iv2]), 
				&(vl->allocatedVectors[cc->iv3]) );
	if ( error[1] < csum.error ) error[1] = csum.error;

	if ( split >= 0 ) {
	  if ( minerror > error[1] ) {
	    split = 1;
	    minerror = error[1];
	  }
	}
	else {
	  split = 1;
	  minerror = error[1];
	}

      }

      if ( ok[2] == 1 ) {

	_init_vector( &sum );
	sum.x = vl->allocatedVectors[cc->iv1].x + vl->allocatedVectors[cc->iv2].x;
	sum.y = vl->allocatedVectors[cc->iv1].y + vl->allocatedVectors[cc->iv2].y;
	sum.z = vl->allocatedVectors[cc->iv1].z + vl->allocatedVectors[cc->iv2].z;
	_compute_3D_uv_vectors( &sum, s );

	_init_cone( &csum );
	_compute_3D_cone_error( &csum, 
				&sum,
				&(vl->allocatedVectors[cc->iv2]), 
				&(vl->allocatedVectors[cc->iv3]) );
	error[2] = csum.error;

	_compute_3D_cone_error( &csum, 
				&(vl->allocatedVectors[cc->iv1]), 
				&sum,
				&(vl->allocatedVectors[cc->iv3]) );
	if ( error[2] < csum.error ) error[2] = csum.error;

	if ( split >= 0 ) {
	  if ( minerror > error[2] ) {
	    split = 2;
	    minerror = error[2];
	  }
	}
	else {
	  split = 2;
	  minerror = error[2];
	}

      }

      
      /* split
       */
      switch ( split ) {

      case 0 :

	iv1 = cc->iv2;
	iv2 = cc->iv3;
	iv3 = cc->iv1;
	
	_split_3D_cone_along_JK( vl, cl, s, i );

	/* look for the cone that shares JK
	 */
	adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
	if ( adjcone < i && adjcone != -1 ) {
	  fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
	}
	
	switch( adjside ) {
	default : break;
	case 0 :
	  _split_3D_cone_along_JK( vl, cl, s, adjcone );
	  break;
	case 1 :
	  _split_3D_cone_along_KI( vl, cl, s, adjcone );
	  break;
	case 2 :
	  _split_3D_cone_along_IJ( vl, cl, s, adjcone );
	  break;
	}
	
	/* consider the cone again  
	 */
	i--;
	
	break;

      case 1 :
	
	iv1 = cc->iv3;
	iv2 = cc->iv1;
	iv3 = cc->iv2;
	
	_split_3D_cone_along_KI( vl, cl, s, i );

	/* look for the cone that shares KI
	 */
	
	adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
	if ( adjcone < i && adjcone != -1 ) {
	  fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
	}
	
	switch( adjside ) {
	default : break;
	case 0 :
	  _split_3D_cone_along_JK( vl, cl, s, adjcone );
	  break;
	case 1 :
	  _split_3D_cone_along_KI( vl, cl, s, adjcone );
	  break;
	case 2 :
	  _split_3D_cone_along_IJ( vl, cl, s, adjcone );
	  break;
	}
	
	/* consider the cone again  
	 */
	i--;
	
	break;

      case 2 :
	
	iv1 = cc->iv1;
	iv2 = cc->iv2;
	iv3 = cc->iv3;
	
	_split_3D_cone_along_IJ( vl, cl, s, i );

	/* look for the cone that shares IJ
	 */
	
	adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
	if ( adjcone < i && adjcone != -1 ) {
	  fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
	}
	
	switch( adjside ) {
	default : break;
	case 0 :
	  _split_3D_cone_along_JK( vl, cl, s, adjcone );
	  break;
	case 1 :
	  _split_3D_cone_along_KI( vl, cl, s, adjcone );
	  break;
	case 2 :
	  _split_3D_cone_along_IJ( vl, cl, s, adjcone );
	  break;
	}
	
	/* consider the cone again  
	 */
	i--;
	
	break;

      } /* end switch */

    }

  }

}



void _build_3D_mask_wrt_error( FILE *fout, VectorList *vl, ConeList *cl, 
			       double *s, int size, int n_vectors,
			       double mask_error ) 
{
  char *proc = "_build_3D_mask_wrt_error";
  int d = (size-1)/2;
  int ok[3];
  double error[3], minerror = 0;
  int split;
  Vector sum;
  Cone *cc, csum;
  int i=0;

  int iv1, iv2, iv3;
  int adjcone, adjside;


  if ( _trace_ ) {
    _print_2D_vector_list( fout, vl );
    _print_cone_list( fout, vl, cl );
  }


  /* loop on cones
   */
  while ( i<cl->n_cones && vl->n_vectors< n_vectors 
	  && cl->allocatedCones[i].error > mask_error ) {
    
    cc = &(cl->allocatedCones[i]);
      
    ok[0] = ok[1] = ok[2] = 1;
    
    /* is it possible to split ?
     */
    if ( vl->allocatedVectors[cc->iv1].x + vl->allocatedVectors[cc->iv2].x > d 
	 || vl->allocatedVectors[cc->iv1].y + vl->allocatedVectors[cc->iv2].y > d
	 || vl->allocatedVectors[cc->iv1].z + vl->allocatedVectors[cc->iv2].z > d )
      ok[2] = 0;
    if ( vl->allocatedVectors[cc->iv2].x + vl->allocatedVectors[cc->iv3].x > d 
	 || vl->allocatedVectors[cc->iv2].y + vl->allocatedVectors[cc->iv3].y > d 
	 || vl->allocatedVectors[cc->iv2].z + vl->allocatedVectors[cc->iv3].z > d )
      ok[0] = 0;
    if ( vl->allocatedVectors[cc->iv3].x + vl->allocatedVectors[cc->iv1].x > d 
	 || vl->allocatedVectors[cc->iv3].y + vl->allocatedVectors[cc->iv1].y > d 
	 || vl->allocatedVectors[cc->iv3].z + vl->allocatedVectors[cc->iv1].z > d )
      ok[1] = 0;
	
    if ( ok[0] == 0 && ok[1] == 0 && ok[2] == 0 ) {
      i++;
      continue;
    }
    
    /* check assumptions
     */
    split = -1;
    if ( ok[0] == 1 ) {

      _init_vector( &sum );
      sum.x = vl->allocatedVectors[cc->iv2].x + vl->allocatedVectors[cc->iv3].x;
      sum.y = vl->allocatedVectors[cc->iv2].y + vl->allocatedVectors[cc->iv3].y;
      sum.z = vl->allocatedVectors[cc->iv2].z + vl->allocatedVectors[cc->iv3].z;
      _compute_3D_uv_vectors( &sum, s );
      
      _init_cone( &csum );
      _compute_3D_cone_error( &csum, 
			      &(vl->allocatedVectors[cc->iv1]), 
			      &sum,
			      &(vl->allocatedVectors[cc->iv3]) );
      error[0] = csum.error;
      
      _compute_3D_cone_error( &csum, 
			      &(vl->allocatedVectors[cc->iv1]), 
			      &(vl->allocatedVectors[cc->iv2]), 
			      &sum );
      if ( error[0] < csum.error ) error[0] = csum.error;
      
      split = 0;
      minerror = error[0];
      
    }
    
    if ( ok[1] == 1 ) {
      
      _init_vector( &sum );
      sum.x = vl->allocatedVectors[cc->iv3].x + vl->allocatedVectors[cc->iv1].x;
      sum.y = vl->allocatedVectors[cc->iv3].y + vl->allocatedVectors[cc->iv1].y;
      sum.z = vl->allocatedVectors[cc->iv3].z + vl->allocatedVectors[cc->iv1].z;
      _compute_3D_uv_vectors( &sum, s );
      
      _init_cone( &csum );
      _compute_3D_cone_error( &csum, 
			      &(vl->allocatedVectors[cc->iv1]), 
			      &(vl->allocatedVectors[cc->iv2]), 
			      &sum );
      error[1] = csum.error;
      
      _compute_3D_cone_error( &csum, 
			      &sum,
			      &(vl->allocatedVectors[cc->iv2]), 
			      &(vl->allocatedVectors[cc->iv3]) );
      if ( error[1] < csum.error ) error[1] = csum.error;
      
      if ( split >= 0 ) {
	if ( minerror > error[1] ) {
	  split = 1;
	  minerror = error[1];
	}
      }
      else {
	split = 1;
	minerror = error[1];
      }
      
    }
    
    if ( ok[2] == 1 ) {
      
      _init_vector( &sum );
      sum.x = vl->allocatedVectors[cc->iv1].x + vl->allocatedVectors[cc->iv2].x;
      sum.y = vl->allocatedVectors[cc->iv1].y + vl->allocatedVectors[cc->iv2].y;
      sum.z = vl->allocatedVectors[cc->iv1].z + vl->allocatedVectors[cc->iv2].z;
      _compute_3D_uv_vectors( &sum, s );
      
      _init_cone( &csum );
      _compute_3D_cone_error( &csum, 
			      &sum,
			      &(vl->allocatedVectors[cc->iv2]), 
			      &(vl->allocatedVectors[cc->iv3]) );
      error[2] = csum.error;
      
      _compute_3D_cone_error( &csum, 
			      &(vl->allocatedVectors[cc->iv1]), 
			      &sum,
			      &(vl->allocatedVectors[cc->iv3]) );
      if ( error[2] < csum.error ) error[2] = csum.error;
      
      if ( split >= 0 ) {
	if ( minerror > error[2] ) {
	  split = 2;
	  minerror = error[2];
	}
      }
      else {
	split = 2;
	minerror = error[2];
      }
      
    }

      
    /* split
     */
    switch ( split ) {
      
    case 0 :
      
      iv1 = cc->iv2;
      iv2 = cc->iv3;
      iv3 = cc->iv1;
      
      _split_3D_cone_along_JK( vl, cl, s, i );
      
      /* look for the cone that shares JK
       */
      adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
      if ( adjcone < i && adjcone != -1 ) {
	fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
      }
      
      switch( adjside ) {
      default : break;
      case 0 :
	_split_3D_cone_along_JK( vl, cl, s, adjcone );
	break;
      case 1 :
	_split_3D_cone_along_KI( vl, cl, s, adjcone );
	break;
      case 2 :
	_split_3D_cone_along_IJ( vl, cl, s, adjcone );
	break;
      }
      
      if ( _trace_ ) {
	fprintf( fout, "add vector (%2d %2d %2d)", 
		 vl->allocatedVectors[iv1].x + vl->allocatedVectors[iv2].x,
		 vl->allocatedVectors[iv1].y + vl->allocatedVectors[iv2].y,
		 vl->allocatedVectors[iv1].z + vl->allocatedVectors[iv2].z );
	fprintf( fout, " => maximal error becomes %f\n", 
		 cl->allocatedCones[0].error );
      }

      break;

    case 1 :
      
      iv1 = cc->iv3;
      iv2 = cc->iv1;
      iv3 = cc->iv2;
      
      _split_3D_cone_along_KI( vl, cl, s, i );
      
      /* look for the cone that shares KI
       */
      
      adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
      if ( adjcone < i && adjcone != -1 ) {
	fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
      }
      
      switch( adjside ) {
      default : break;
      case 0 :
	_split_3D_cone_along_JK( vl, cl, s, adjcone );
	break;
      case 1 :
	_split_3D_cone_along_KI( vl, cl, s, adjcone );
	break;
      case 2 :
	_split_3D_cone_along_IJ( vl, cl, s, adjcone );
	break;
      }
      
      if ( _trace_ ) {
	fprintf( fout, "add vector (%2d %2d %2d)", 
		 vl->allocatedVectors[iv1].x + vl->allocatedVectors[iv2].x,
		 vl->allocatedVectors[iv1].y + vl->allocatedVectors[iv2].y,
		 vl->allocatedVectors[iv1].z + vl->allocatedVectors[iv2].z );
	fprintf( fout, " => maximal error becomes %f\n", 
		 cl->allocatedCones[0].error );
      }

      break;
      
    case 2 :
      
      iv1 = cc->iv1;
      iv2 = cc->iv2;
      iv3 = cc->iv3;
      
      _split_3D_cone_along_IJ( vl, cl, s, i );
      
      /* look for the cone that shares IJ
       */
      
      adjcone = _search_adjacent_3D_cone( cl, iv1, iv2, iv3, &adjside );
      if ( adjcone < i && adjcone != -1 ) {
	fprintf( stderr, "%s: ERROR: this case should not occur\n", proc );
      }
      
      switch( adjside ) {
      default : break;
      case 0 :
	_split_3D_cone_along_JK( vl, cl, s, adjcone );
	break;
      case 1 :
	_split_3D_cone_along_KI( vl, cl, s, adjcone );
	break;
      case 2 :
	_split_3D_cone_along_IJ( vl, cl, s, adjcone );
	break;
      }
      
      if ( _trace_ ) {
	fprintf( fout, "add vector (%2d %2d %2d)", 
		 vl->allocatedVectors[iv1].x + vl->allocatedVectors[iv2].x,
		 vl->allocatedVectors[iv1].y + vl->allocatedVectors[iv2].y,
		 vl->allocatedVectors[iv1].z + vl->allocatedVectors[iv2].z );
	fprintf( fout, " => maximal error becomes %f\n", 
		 cl->allocatedCones[0].error );
      }

      break;
      
    } /* end switch */
    
  }

}



void _sort_3D_vectors( Vector **v, VectorList *vl, ConeList *cl )
{
  char *proc = "_sort_3D_vectors";
  int i, j, n;
  int t;
  int *tab = NULL;

  for ( i=0; i<vl->n_vectors; i++ ) 
    vl->allocatedVectors[i].flag = 0;
  for ( i=0; i<cl->n_cones; i++ ) 
    cl->allocatedCones[i].flag = 0;


  tab = (int*)malloc( vl->n_vectors * sizeof(int) );
  if ( tab == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    exit( 0 );
  }
  /* first cone
   */

  switch( cl->allocatedCones[0].type ) {
  case _UNSET_ :
    fprintf( stderr, "%s: the first cone has no type\n", proc );
  case _INSIDE_ :
  if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) )
       < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv3]) ) ) {
    if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	 < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) ) ) {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
      v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
    }
    else {
      if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	   < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv3]) ) ) {
	v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
	v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
	v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
      }
      else {
	v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
	v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
	v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      }
    }
  }
  else {
    if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	 < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv3]) ) ) {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
      v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
    }
    else {
      if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	   < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) ) ) {
	v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
	v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
	v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
      }
      else {
	v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
	v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
	v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      }
    }
  }
  break;
  case _OUT_IJ_ :
    if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	 < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) ) ) {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
    }
    else {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
    }
    v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
    break;
  case _OUT_JK_ :
    if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv2]) )
	 < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv3]) ) ) {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
    }
    else {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
    }
    v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
    break;
  case _OUT_KI_ :
    if ( _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]) ) 
	 < _norme2v_3D( &(vl->allocatedVectors[cl->allocatedCones[0].iv3]) ) ) {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
    }
    else {
      v[0] = &(vl->allocatedVectors[cl->allocatedCones[0].iv3]);
      v[1] = &(vl->allocatedVectors[cl->allocatedCones[0].iv1]);
    }
    v[2] = &(vl->allocatedVectors[cl->allocatedCones[0].iv2]);
    break;
  }

  cl->allocatedCones[0].flag = 1;
  v[0]->flag = 1;
  v[1]->flag = 1;
  v[2]->flag = 1;

  n = 3;


  /* let's go for the other vectors
   */

  for ( ; n < vl->n_vectors;  ) {

    t = 0;
    /* find adjacents cones 
     */

    for ( i=1; i<cl->n_cones; i++ ) {

    /* skip marked cones
     */

      if ( cl->allocatedCones[i].flag == 1 ) continue;
      if ( vl->allocatedVectors[cl->allocatedCones[i].iv1].flag == 1 &&
	   vl->allocatedVectors[cl->allocatedCones[i].iv2].flag == 1 &&
	   vl->allocatedVectors[cl->allocatedCones[i].iv3].flag == 1 ) {
	cl->allocatedCones[i].flag = 1;
	continue;
      }

      /* try to find a cone with two marked vertices
	 ie with only one unmarked vertex
      */
      
      if ( (vl->allocatedVectors[cl->allocatedCones[i].iv1].flag == 0 
	    && vl->allocatedVectors[cl->allocatedCones[i].iv2].flag == 0)
	   || (vl->allocatedVectors[cl->allocatedCones[i].iv2].flag == 0 
	       && vl->allocatedVectors[cl->allocatedCones[i].iv3].flag == 0)
	   || (vl->allocatedVectors[cl->allocatedCones[i].iv3].flag == 0 
	       && vl->allocatedVectors[cl->allocatedCones[i].iv1].flag == 0) )
	continue;

      tab[t] = i;
      t++;
    }

    /* select the cone from the candidates
     */
    switch ( t ) {
    case 0 :
      /* this should not occur
       */
      j = 0 ;
      break;
    case 1 :
      /* take the first cone in the list
       */
      j = tab[0];
      break;
    default :
      /* take the cone of minimal error in the list
       */
      for ( j=tab[0], i=1; i<t; i++ ) 
	if ( cl->allocatedCones[j].error > cl->allocatedCones[tab[i]].error )
	  j = tab[i];
    }
    

    /* add the unmarked vertex into the sorted
       list of vectors
    */

    if ( j > 0 ) {
      cl->allocatedCones[j].flag = 1;
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv1].flag == 0 ) {
	v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv1]);
	v[n]->flag = 1;
	n++;
      }
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv2].flag == 0 ) {
	v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv2]);
	v[n]->flag = 1;
	n++;
      }
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv3].flag == 0 ) {
	v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv3]);
	v[n]->flag = 1;
	n++;
      }
      continue;
    }

    /* do not find a cone with two vectors already in list
     */
    
    fprintf( stderr, "%s: ERROR, looking for a cone with one marked vertex\n", 
	     proc );
    
    for ( j=0, i=1; i<cl->n_cones && j==0; i++ ) {
      if ( cl->allocatedCones[i].flag == 1 ) continue;
      if ( vl->allocatedVectors[cl->allocatedCones[i].iv1].flag == 0 
	   && vl->allocatedVectors[cl->allocatedCones[i].iv2].flag == 0 
	   && vl->allocatedVectors[cl->allocatedCones[i].iv3].flag == 0 )
	continue;
      j = i;
    }
    
    if ( j > 0 ) {

      if ( vl->allocatedVectors[cl->allocatedCones[j].iv1].flag == 1 ) {
	if ( vl->allocatedVectors[cl->allocatedCones[j].iv2].vn 
	     < vl->allocatedVectors[cl->allocatedCones[j].iv3].vn ) {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv2]);
	  v[n]->flag = 1;
	  n++;
	}
	else {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv3]);
	  v[n]->flag = 1;
	  n++;
	}
      }
      
      if ( vl->allocatedVectors[cl->allocatedCones[j].iv2].flag == 1 ) {
	if ( vl->allocatedVectors[cl->allocatedCones[j].iv1].vn 
	     < vl->allocatedVectors[cl->allocatedCones[j].iv3].vn ) {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv1]);
	  v[n]->flag = 1;
	  n++;
	}
	else {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv3]);
	  v[n]->flag = 1;
	  n++;
	}
      }
      
       if ( vl->allocatedVectors[cl->allocatedCones[j].iv3].flag == 1 ) {
	if ( vl->allocatedVectors[cl->allocatedCones[j].iv1].vn 
	     < vl->allocatedVectors[cl->allocatedCones[j].iv2].vn ) {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv1]);
	  v[n]->flag = 1;
	  n++;
	}
	else {
	  v[n] = &(vl->allocatedVectors[cl->allocatedCones[j].iv2]);
	  v[n]->flag = 1;
	  n++;
	}
      }
    }

  }

  free( tab );
}



void _rearrange_3D_vectors_in_cone( VectorList *vl, ConeList *cl )
{
  int i, t;
  double d;
  Vector *v1, *v2, *v3;

  for ( i=0; i<cl->n_cones; i++ ) {
    v1 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv1 ]);
    v2 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv2 ]);
    v3 = &(vl->allocatedVectors[ cl->allocatedCones[i].iv3 ]);
    d = v1->vx * v2->vy * v3->vz
      + v1->vy * v2->vz * v3->vx
      + v1->vz * v2->vx * v3->vy
      - v1->vx * v2->vz * v3->vy
      - v1->vy * v2->vx * v3->vz
      - v1->vz * v2->vy * v3->vx;
    if ( d < 0 ) {
      t = cl->allocatedCones[i].iv2;
      cl->allocatedCones[i].iv2 = cl->allocatedCones[i].iv3;
      cl->allocatedCones[i].iv3 = t;
    }
  }
}



void _compute_3D_weights( FILE *fout, ResultList *rl, 
			  VectorList *vl, ConeList *cl,
			  int f, int l ) 
{
  char *proc = "_compute_3D_weights";
  Vector **v;
  int i, j;
  double dmin, dmax, eps;
  double perr=1.0;
  double erropt, err;
  int iv;

  double v1min, v1max;

  int ok;
  Cone * cone;
  double ldmin, ldmax, lerr, leps;

  double v12[3], v23[3], v31[3];
  double v12_12, v12_23, v12_31, v23_23, v23_31, v31_31;

  struct tms timestruct;
  double it, ct;



  /* add links from vectors to cones
   */
  (void)_link_cones_to_vectors( vl, cl );

  /* allocate and sort the vectors
   */
  v = (Vector**)malloc( vl->n_vectors * sizeof(Vector*) );
  if ( v == NULL ) {
    fprintf( stderr, "%s: allocation error\n", proc );
    return;
  }
  _sort_3D_vectors( v, vl, cl );
  if ( _trace_ || _debug_ ) {
    for ( i=0; i<vl->n_vectors; i++ ) {
      fprintf( fout, "v[#%3d] = ", i );
      _print_3D_vector( fout, v[i] );
      fprintf( fout, "\n" );
    }
  }

  /* initialization
     flag = 0 -> no set yet
     flag = 1 -> already set
  */

  for ( i=0; i<vl->n_vectors; i++ ) 
    vl->allocatedVectors[i].flag = 0;
  for ( i=0; i<cl->n_cones; i++ ) 
    cl->allocatedCones[i].flag = 0;


  /* some vectorial and scalar products
   */
  v23[0] = v[1]->vy * v[2]->vz - v[1]->vz * v[2]->vy;
  v23[1] = v[1]->vz * v[2]->vx - v[1]->vx * v[2]->vz;
  v23[2] = v[1]->vx * v[2]->vy - v[1]->vy * v[2]->vx;

  v31[0] = v[2]->vy * v[0]->vz - v[2]->vz * v[0]->vy;
  v31[1] = v[2]->vz * v[0]->vx - v[2]->vx * v[0]->vz;
  v31[2] = v[2]->vx * v[0]->vy - v[2]->vy * v[0]->vx;

  v12[0] = v[0]->vy * v[1]->vz - v[0]->vz * v[1]->vy;
  v12[1] = v[0]->vz * v[1]->vx - v[0]->vx * v[1]->vz;
  v12[2] = v[0]->vx * v[1]->vy - v[0]->vy * v[1]->vx;
  
  v12_12 = v12[0]*v12[0] + v12[1]*v12[1] + v12[2]*v12[2];
  v12_23 = v12[0]*v23[0] + v12[1]*v23[1] + v12[2]*v23[2];
  v12_31 = v12[0]*v31[0] + v12[1]*v31[1] + v12[2]*v31[2];

  v23_23 = v23[0]*v23[0] + v23[1]*v23[1] + v23[2]*v23[2];
  v23_31 = v23[0]*v31[0] + v23[1]*v31[1] + v23[2]*v31[2];

  v31_31 = v31[0]*v31[0] + v31[1]*v31[1] + v31[2]*v31[2];


  /* the largest error
   */
  erropt = cl->allocatedCones[0].error;


  /* set bounds for the vectors of the first cone
     low bound = f
     high bound = l * |v|_1
   */
  v[0]->f = floor(f * _normeIv_3D( v[0] ) );
  v[0]->l = (int)(l * _norme1v_3D( v[0] ) );

  v[1]->f = floor(f * _normeIv_3D( v[1] ) );
  v[1]->l = (int)(l * _norme1v_3D( v[1] ) );

  v[2]->f = floor(f * _normeIv_3D( v[2] ) );
  v[2]->l = (int)(l * _norme1v_3D( v[2] ) );

  v[0]->flag = 1;
  v[1]->flag = 1;
  v[2]->flag = 1;


  it = times( &timestruct );



  /* loops for the first cone
   */
  for ( v[0]->w  = v[0]->f;
	v[0]->w <= v[0]->l;
	v[0]->w ++ ) {

    if ( _trace_ )
      fprintf( fout,"%3d   [%d %d]\n", v[0]->w, v[0]->f, v[0]->l );



    /* get better bounds for the second vector
       to satisfy the 2D convexity conditions
    */
    if ( _criterium_ == _MAXIMALITY_ ) {
      v1min = v[0]->w 
	* (v[0]->vx * v[1]->vx + v[0]->vy * v[1]->vy + v[0]->vz * v[1]->vz)
	/ (v[0]->vx * v[0]->vx + v[0]->vy * v[0]->vy + v[0]->vz * v[0]->vz);
      v1max = v[0]->w
	* (v[1]->vx * v[1]->vx + v[1]->vy * v[1]->vy + v[1]->vz * v[1]->vz)
	/ (v[0]->vx * v[1]->vx + v[0]->vy * v[1]->vy + v[0]->vz * v[1]->vz);
      
      v[1]->f = ceil( v1min );
      v[1]->l = floor( v1max );
      if ( v[1]->f <= 0 ) v[1]->f = 1;
    }

    for ( v[1]->w  = v[1]->f;
	  v[1]->w <= v[1]->l;
	  v[1]->w ++ ) {
      
      if ( _trace_ )
	fprintf( fout,"%3d %3d   [%d %d]\n", v[0]->w, v[1]->w, v[1]->f, v[1]->l );
      
      
      /* get better bounds for the third vector
	 to satisfy the 2D convexity conditions
      */
      if ( _criterium_ == _MAXIMALITY_ ) {
	v[2]->f = f;
	v[2]->l = (int)(l * _norme1v_3D( v[2] ) );
	v[2]->f = ceil( - (v[0]->w * v12_23 + v[1]->w * v12_31) / v12_12 );
	if ( v12_31 > 0 ) {
	  if ( v[2]->f < ceil( - (v[0]->w * v23_31 + v[1]->w * v31_31) / v12_31 ) )
	    v[2]->f = ceil( - (v[0]->w * v23_31 + v[1]->w * v31_31) / v12_31 );
	}
	else if ( v12_31 < 0 ) {
	  if ( v[2]->l > floor((v[0]->w * v23_31 + v[1]->w * v31_31) / (-v12_31) ) )
	    v[2]->l = floor((v[0]->w * v23_31 + v[1]->w * v31_31) / (-v12_31));
	}
	if ( v12_23 > 0 ) {
	  if ( v[2]->f < ceil( - (v[0]->w * v23_23 + v[1]->w * v23_31) / v12_23 ) )
	    v[2]->f = ceil( - (v[0]->w * v23_23 + v[1]->w * v23_31) / v12_23 );
	}
	else if ( v12_23 < 0 ) {
	  if ( v[2]->l > floor((v[0]->w * v23_23 + v[1]->w * v23_31) / (-v12_23) ) )
	    v[2]->l = floor((v[0]->w * v23_23 + v[1]->w * v23_31) / (-v12_23));
	}
      }
      
      
  
      for ( v[2]->w  = v[2]->f;
	    v[2]->w <= v[2]->l;
	    v[2]->w ++ ) {

	if ( _trace_ )
	  fprintf( fout,"%3d %3d %3d   [%d %d]\n", 
		   v[0]->w, v[1]->w, v[2]->w, v[2]->f, v[2]->l );

	/* check condition
	 */
	switch( _criterium_ ) {
	case _MAXIMALITY_ :
	  if ( _check_3D_maximality( vl, &(cl->allocatedCones[0]) ) == 0 ) {
	    if ( _trace_ ) 
	      fprintf( fout," cone #1 not convex\n" );
	    continue; 
	  }
	  break;
	case _CONVEXITY_ :
	  if ( _check_3D_convexity( vl, cl, &(cl->allocatedCones[0]) ) == 0 ) {
	    if ( _trace_ )
	      fprintf( fout, " cone #1 not convex\n" );
	    continue; 
	  }
	  break;   
	}
	
	/* compute min and max distance, epsilon and error
	 */      
	_compute_3D_cone_min_max( &(vl->allocatedVectors[cl->allocatedCones[0].iv1]),
				  &(vl->allocatedVectors[cl->allocatedCones[0].iv2]),
				  &(vl->allocatedVectors[cl->allocatedCones[0].iv3]),
				  &dmin, &dmax );

	eps = (dmin + dmax)/2.0;
	err = dmax/eps - 1;

	/* if we do not perform better,
	   go to next
	*/
	if ( err >= perr ) {
	  if ( _trace_ ) 
	    fprintf( fout,"%3d %3d %3d too large error\n", v[0]->w, v[1]->w, v[2]->w );
	  continue;
	}
	
	/* Here, we've got better candidates
	   for the cone of largest error
	*/

	/* special case:
	   there is only one cone
	*/
	iv = 3;
	if ( iv == vl->n_vectors ) {
	  ct = times( &timestruct );
	  _add_result_to_list( rl, vl,  err, dmin, dmax, eps, ct, it );
	  _print_cone_information( fout, erropt, perr, err, dmin, dmax, eps, ct, it );
	  _print_3D_weighted_vector_list( fout, vl, 3 );
	  perr = err;
	  continue;
	}
	
	if ( _verbose_ >= 3 ) {
	  fprintf( fout, "\n" );
	  fprintf( fout, "* v[0]=%d in [%d %d] v[1]=%d in [%d %d] v[2]=%d in [%d %d]\n", 
		   v[0]->w, v[0]->f, v[0]->l, 
		   v[1]->w, v[1]->f, v[1]->l, 
		   v[2]->w, v[2]->f, v[2]->l );
	}

	cl->allocatedCones[0].flag = 1;

	/* here we have more work to do
	   compute the bounds for the other vectors
	   from the error of the first cone
	*/
	for ( i=3, ok=1; i<vl->n_vectors && ok==1; i++ ) {

	  if ( _be_strict_ == 0 ) {
	    v[i]->f = (int)(floor( eps*(1-perr)*v[i]->vn ) + 0.5 );
	    v[i]->l = (int)(ceil(  eps*(1+perr)*v[i]->vn ) + 0.5 );
	  }
	  else {
	    v[i]->f = (int)(ceil(  eps*(1-perr)*v[i]->vn ) + 0.5 );
	    v[i]->l = (int)(floor( eps*(1+perr)*v[i]->vn ) + 0.5 );
	  }

	  if ( v[i]->f <= 0 ) v[i]->f = 1;
	  if ( v[i]->l > (int)(l * _norme1v_3D(v[i])) ) 
	    v[i]->l = (int)(l * _norme1v_3D(v[i]));

	  v[i]->w = v[i]->f-1;

	  if ( _verbose_ >= 3 ) {
	    fprintf( stdout, "  bounds #%2d: %3d -> %3d\n", i,
		     v[i]->f, v[i]->l );
	  }
	  if ( v[i]->l < v[i]->f ) ok = 0;
	}

	if ( ok == 0 ) {
	  if ( _trace_ ) 
	    fprintf( fout,"%3d %3d %3d empty interval\n", v[0]->w, v[1]->w, v[2]->w );
	  if ( _verbose_ >= 3 ) {
	    fprintf( stdout, "  empty interval\n" );
	  }
	  continue;
	}


	
	
	
	/* bounds have been computed
	   we have to find a weight set that is optimal
	*/
	do {
	  
	  /* find a weight that satisfy convexity conditions
	     for next vector
	  */
	  do {
	    v[iv]->w ++;
	    if ( _trace_ ) {
	      if ( v[iv]->w == v[iv]->f ) {
		fprintf( fout,"%3d %3d %3d", v[0]->w, v[1]->w, v[2]->w );
		for (i=3; i<=iv; i++ )
		  fprintf( fout," %3d", v[i]->w );
		fprintf( fout,"   [%d %d]\n", v[iv]->f, v[iv]->l );
	      }
	    }
	    if ( v[iv]->w <= v[iv]->l ) {
	      v[iv]->flag = 1;
	      ok = 1;
	      for ( j = 0; j<v[iv]->n_cones && ok == 1; j++ ) {
		cone = v[iv]->c[j];
		if ( _criterium_ == _MAXIMALITY_ && cone->flag == 1 ) continue;
		if ( (&(vl->allocatedVectors[cone->iv1]) == v[iv] 
		      && vl->allocatedVectors[cone->iv2].flag == 1
		      && vl->allocatedVectors[cone->iv3].flag == 1)
		     || (vl->allocatedVectors[cone->iv1].flag == 1 
			 && &(vl->allocatedVectors[cone->iv2]) == v[iv] 
			 && vl->allocatedVectors[cone->iv3].flag == 1)
		     || (vl->allocatedVectors[cone->iv1].flag == 1 
			 && vl->allocatedVectors[cone->iv2].flag == 1
			 && &(vl->allocatedVectors[cone->iv3]) == v[iv]) ) {
		  switch( _criterium_ ) {
		  case _MAXIMALITY_ :
		    ok = ok &&  _check_3D_maximality( vl, cone );
		    break;
		  case _CONVEXITY_ :
		    ok = ok && _check_3D_convexity( vl, cl, cone );
		    break;   
		  }
		  if ( _trace_ && ok == 0 ) {
		    fprintf( fout,"%3d %3d %3d", v[0]->w, v[1]->w, v[2]->w );
		    for (i=3; i<=iv; i++ )
		      fprintf( fout," %3d", v[i]->w );
		    fprintf( fout,"   not maximal/convex\n" );
		  }
		  if ( _criterium_ == _MAXIMALITY_ && ok == 1 ) cone->flag = 1;
		}
	      }
	    }
	    
	  } while ( v[iv]->w <= v[iv]->l && ok == 0 );
	  
	  
	  /* if not possible
	     get back to previous vector
	  */
	  if ( v[iv]->w > v[iv]->l ) {
	    if ( _verbose_ >= 3 ) {
	      fprintf( stdout, "   TEST FAILURE " );
	      for (j=3; j<iv; j++ ) {
		fprintf( stdout, "#%d=%d ", j, v[j]->w );
	      }
	      fprintf( stdout, "#%d=\n", iv );
	    }
	    v[iv]->w = v[iv]->f-1;
	    for ( j = 0; j<v[iv]->n_cones; j++ )
	      v[iv]->c[j]->flag = 0;
	    v[iv]->flag = 0;
	    iv --;
	  }
	  /* or get back to next
	   */
	  else {
	    if ( iv == vl->n_vectors-1 ) {
	      /* we got a mask
	       */
	      if ( _verbose_ >= 3 ) {
		fprintf( stdout, "   TEST SUCCESS " );
		for (j=2; j<=iv; j++ ) {
		  fprintf( stdout, "#%d=%d ", j, v[j]->w );
		}
		fprintf( stdout, "\n" );
	      }
	      _compute_3D_mask_min_max( vl, cl, &ldmin, &ldmax );
	      leps = (ldmin + ldmax)/2.0;
	      lerr = ldmax/leps - 1;
	      
	      if ( 1.0001 * lerr < perr ) {
		ct = times( &timestruct );
		_add_result_to_list( rl, vl,  lerr, ldmin, ldmax, leps, ct, it );
		_print_mask_information( fout, erropt, perr, 
					 err, dmin, dmax, eps,
					 lerr, ldmin, ldmax, leps, ct, it );
		_print_3D_weighted_vector_list( fout, vl, vl->n_vectors );
		perr = lerr;
	      }
	    }
	    else {
	      v[iv]->flag = 1;
	      iv ++;
	    }
	  }
	  
	  
	} while ( iv > 2 );





      
      } /* for v[2] */
    }   /* for v[0] */
  }     /* for v[0] */

  free(v);
}








/*==================================================
 *
 * MATLAB
 *
 ==================================================*/



void _print_2D_mask_matlab( FILE *f, VectorList *vl, ConeList *cl )
{
  int i;
  double emin, emax;
  double x1, y1, n1;
  double x2, y2, n2;

  emin = emax = cl->allocatedCones[0].error;
  for ( i=1; i<cl->n_cones; i++ ) {
    if ( emin > cl->allocatedCones[i].error )
      emin = cl->allocatedCones[i].error;
    if ( emax < cl->allocatedCones[i].error )
      emax = cl->allocatedCones[i].error;
  }
  if ( emax == emin ) emax += 0.01;
  
  fprintf( f, "\n" );
  fprintf( f, "emin=%f;\n", emin );
  fprintf( f, "emax=%f;\n", emax );
  fprintf( f, "\n" );

  fprintf( f, "figure\n" );
  fprintf( f, "hold on\n" );
  fprintf( f, "\n" );
  fprintf( f, "col=colormap(jet);\n" );
  fprintf( f, "cmin=1;\n" );
  fprintf( f, "cmax=size(col,1);\n" );
  fprintf( f, "alpha=0.75;\n");
  fprintf( f, "\n" );
  

  fprintf( f, "if (1)\n" );
  for (i=0; i<vl->n_vectors; i++ ) {
    if ( 1 ) {
      x1 = vl->allocatedVectors[i].x;
      y1 = vl->allocatedVectors[i].y;
      n1 = sqrt( x1*x1 + y1*y1 );
      fprintf( f, "  plot([0,%f],[0,%f], 'k', 'LineWidth', 2 );\n", 
	       x1/n1, y1/n1 );
    }
    else {
      fprintf( f, "  plot([0,%f],[0,%f], 'k', 'LineWidth', 2 );\n", 
	       vl->allocatedVectors[i].ux, vl->allocatedVectors[i].uy );
    }
  }
  fprintf( f, "end\n" );
  fprintf( f, "\n" );
    
  for ( i=0; i<cl->n_cones; i++ ) {

    if ( 0 ) 
      fprintf( f, "  ic=cmin+round( (%f-emin)*(cmax-cmin)/(emax-emin) );\n",
	       cl->allocatedCones[i].error );
    fprintf( f, "  ic=cmin+round( (log(%f)-log(emin))*(cmax-cmin)/(log(emax)-log(emin)) );\n",
	     cl->allocatedCones[i].error );
    if ( 0 ) 
      fprintf( stderr, "color %f -> %f\n",
	       1+(cl->allocatedCones[i].error-emin)*(64-1)/(emax-emin),
	       1+(log(cl->allocatedCones[i].error)-log(emin))*(64-1)/(log(emax)-log(emin)) );
	       
     


    if ( 1 ) {
      x1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].x;
      y1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].y;
      x2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].x;
      y2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].y;
      n1 = sqrt( x1*x1 + y1*y1 );
      n2 = sqrt( x2*x2 + y2*y2 );
      fprintf( f, "  p=patch( [0 %f %f 0], [0 %f %f 0], col(ic,:) );\n",
	       x1/n1, x2/n2, y1/n1, y2/n2 );
    }
    else {
      fprintf( f, "  p=patch( [0 %f %f 0], [0 %f %f 0], col(ic,:) );\n",
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].ux, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv2].ux, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].uy, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv2].uy );
    }
    fprintf( f, "  set(p, 'FaceAlpha', alpha );\n" );
  }
  fprintf( f, "\n" );
  fprintf( f, "axis equal;\n" );
  fprintf( f, "axis( [0 1 0 1] );\n" );
  fprintf( f, "view(2);\n" );
  fprintf( f, "grid;\n" );
  fprintf( f, "cb = colorbar;\n" );
  if ( 0 ) 
    fprintf( f, "ct = 0.01*round(10000*(emin:(emax-emin)/10:emax));\n" );
  fprintf( f, "ct = 0.01*round(10000*exp(log(emin):(log(emax)-log(emin))/10:log(emax)));\n" );
  fprintf( f, "set(cb,'YTickLabel',num2str(ct') );\n" );
  fprintf( f, "if (0)\n" );
  fprintf( f, "  print( gcf, '-djpeg85', 'figure.jpg' );\n" );
    fprintf( f, "end\n" );
  fprintf( f, "\n" );
}



void _print_3D_mask_matlab( FILE *f, VectorList *vl, ConeList *cl )
{
  int i;
  double emin, emax;
  double x1, y1, z1, n1;
  double x2, y2, z2, n2;
  double x3, y3, z3, n3;
  
  emin = emax = cl->allocatedCones[0].error;
  for ( i=1; i<cl->n_cones; i++ ) {
    if ( emin > cl->allocatedCones[i].error )
      emin = cl->allocatedCones[i].error;
    if ( emax < cl->allocatedCones[i].error )
      emax = cl->allocatedCones[i].error;
  }
  if ( emax == emin ) emax += 0.01;



  fprintf( f, "\n" );
  fprintf( f, "alpha=0.90;\n");
  fprintf( f, "\n" );
  
  /* to print inner and outer normals
   */
  if ( 0 ) { 

    fprintf( f, "in='b';\n");
    fprintf( f, "ou='r';\n");
    fprintf( f, "un='k';\n");
    fprintf( f, "\n" );
    
    fprintf( f, "figure\n" );
    fprintf( f, "hold on\n" );
    for ( i=0; i<cl->n_cones; i++ ) {
      if ( 1 ) {
	x1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].x;
	y1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].y;
	z1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].z;
	x2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].x;
	y2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].y;
	z2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].z;
	x3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].x;
	y3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].y;
	z3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].z;
	n1 = sqrt( x1*x1 + y1*y1 + z1*z1 );
	n2 = sqrt( x2*x2 + y2*y2 + z2*z2 );
	n3 = sqrt( x3*x3 + y3*y3 + z3*z3 );
	switch( cl->allocatedCones[i].type ) {
	case _UNSET_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], un );\n",
		   x1/n1, x2/n2, x3/n3, x1/n1,
		   y1/n1, y2/n2, y3/n3, y1/n1,
		   z1/n1, z2/n2, z3/n3, z1/n1 );
	  break;
	case _INSIDE_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], in );\n",
		   x1/n1, x2/n2, x3/n3, x1/n1,
		   y1/n1, y2/n2, y3/n3, y1/n1,
		   z1/n1, z2/n2, z3/n3, z1/n1 );
	  break;
	case _OUT_IJ_ :
	case _OUT_JK_ :
	case _OUT_KI_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], ou );\n",
		   x1/n1, x2/n2, x3/n3, x1/n1,
		   y1/n1, y2/n2, y3/n3, y1/n1,
		   z1/n1, z2/n2, z3/n3, z1/n1 );
	  break;
	}
      }
      else {
	switch( cl->allocatedCones[i].type ) {
	case _UNSET_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], un );\n",
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz );
	  break;
	case _INSIDE_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], in );\n",
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz );
	  break;  
	case _OUT_IJ_ :
	case _OUT_JK_ :
	case _OUT_KI_ :
	  fprintf( f, "  p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], ou );\n",
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].ux, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].ux,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uy, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uy,
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv2].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv3].uz, 
		   vl->allocatedVectors[cl->allocatedCones[i].iv1].uz );
	  break;  
	}
      }
      fprintf( f, "  set(p, 'FaceAlpha', alpha );\n" );
    }
    fprintf( f, "\n" );
    fprintf( f, "axis( [0 1 0 1 0 1] );\n" );
    fprintf( f, "%% axis equal;\n" );
    fprintf( f, "view( [110 20]);\n" );
    fprintf( f, "grid;\n" );
    fprintf( f, "if (0)\n" );
    fprintf( f, "  print( gcf, '-djpeg85', 'figure.jpg' );\n" );
    fprintf( f, "end\n" );
  }

  
  
  if ( 0 ) {
    fprintf( f, "if (0)\n" );
    fprintf( f, "  figure\n" );
    fprintf( f, "  hold on\n" );
    fprintf( f, "  set(gca, 'FontWeight','bold','FontSize', 14 );\n" );
    fprintf( f, "\n" );
    fprintf( f, "  colp = [0.5 0.5 0.5];\n" );
    fprintf( f, "\n" );
    for ( i=0; i<cl->n_cones; i++ ) {
      
      if ( 1 ) {
	x1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].x;
	y1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].y;
	z1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].z;
	x2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].x;
	y2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].y;
	z2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].z;
	x3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].x;
	y3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].y;
	z3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].z;
	n1 = sqrt( x1*x1 + y1*y1 + z1*z1 );
	n2 = sqrt( x2*x2 + y2*y2 + z2*z2 );
	n3 = sqrt( x3*x3 + y3*y3 + z3*z3 );
	fprintf( f, "    p=patch( [0 %f %f 0], [0 %f %f 0], [0 %f %f 0], colp );\n",
		 x2/n2, x3/n3, y2/n2, y3/n3, z2/n2, z3/n3 );
	fprintf( f, "%%    set(p, 'FaceAlpha', alpha );\n" );
	fprintf( f, "    p=patch( [0 %f %f 0], [0 %f %f 0], [0 %f %f 0], colp );\n",
		 x3/n3, x1/n1, y3/n3, y1/n1, z3/n3, z1/n1 );
	fprintf( f, "%%    set(p, 'FaceAlpha', alpha );\n" );
	fprintf( f, "    p=patch( [0 %f %f 0], [0 %f %f 0], [0 %f %f 0], colp );\n",
		 x1/n1, x2/n2, y1/n1, y2/n2, z1/n1, z2/n2 );
	fprintf( f, "%%    set(p, 'FaceAlpha', alpha );\n" );
      }
      
    }
    fprintf( f, "\n" );
    for (i=0; i<vl->n_vectors; i++ ) {
      if ( 1 ) {
	x1 = vl->allocatedVectors[i].x;
	y1 = vl->allocatedVectors[i].y;
	z1 = vl->allocatedVectors[i].z;
	n1 = sqrt( x1*x1 + y1*y1 + z1*z1 );
	fprintf( f, "    plot3([%f %f],[%f %f],[%f %f], 'k', 'LineWidth', 3 );\n", 
		 x1/n1, 1.1 * x1/n1, y1/n1, 1.1 * y1/n1, z1/n1, 1.1 * z1/n1 );
      }
    }
    fprintf( f, "\n" );
    fprintf( f, "  if (1)\n" );
    fprintf( f, "    axis( [0 1.1 0 1.1 0 1.1] );\n" );
    fprintf( f, "  else\n" );
    fprintf( f, "    axis( [0.4 1 0 0.8 0 0.6] );\n" );
    fprintf( f, "  end\n" );
    fprintf( f, "\n" );
    fprintf( f, "  %% axis equal;\n" );
    fprintf( f, "  view( [110 20]);\n" );
    fprintf( f, "  grid;\n" );
    fprintf( f, "\n" );
    fprintf( f, "  if (0)\n" );
    fprintf( f, "    print( gcf, '-djpeg85', 'figure.jpg' );\n" );
    fprintf( f, "  end\n" );
    fprintf( f, "\n" );
    fprintf( f, "  hold off\n" );
    fprintf( f, "end\n" );
    fprintf( f, "\n" );
  }




  fprintf( f, "\n" );
  fprintf( f, "emin=%f;\n", emin );
  fprintf( f, "emax=%f;\n", emax );
  fprintf( f, "\n" );

  fprintf( f, "figure\n" );
  fprintf( f, "hold on\n" );
  fprintf( f, "set(gca, 'FontWeight','bold','FontSize', 14 );\n" );
  fprintf( f, "\n" );
  fprintf( f, "col=colormap(jet);\n" );
  fprintf( f, "cmin=1;\n" );
  fprintf( f, "cmax=size(col,1);\n" );
  fprintf( f, "\n" );
  fprintf( f, "cb = colorbar;\n" );
  if ( 0 ) 
    fprintf( f, "ct = 0.01*round(10000*(emin:(emax-emin)/10:emax));\n" );
  fprintf( f, "ct = 0.01*round(10000*exp(log(emin):(log(emax)-log(emin))/5:log(emax)));\n" );
  fprintf( f, "set(cb,'YTick',cmin+1:(cmax-cmin-2)/5:cmax-1);\n" );

  fprintf( f, "set(cb,'YTickLabel',num2str(ct'), 'FontWeight','bold','FontSize', 24 );\n" );
  fprintf( f, "\n" );

  fprintf( f, "if (0)\n" );
  for (i=0; i<vl->n_vectors; i++ ) {
    if ( 1 ) {
      x1 = vl->allocatedVectors[i].x;
      y1 = vl->allocatedVectors[i].y;
      z1 = vl->allocatedVectors[i].z;
      n1 = sqrt( x1*x1 + y1*y1 + z1*z1 );
      fprintf( f, "  plot3([0,%f],[0,%f],[0,%f], 'k', 'LineWidth', 2 );\n", 
	       x1/n1, y1/n1, z1/n1 );
    }
    else {
      fprintf( f, "  plot3([0,%f],[0,%f],[0,%f], 'k', 'LineWidth', 2 );\n", 
	       vl->allocatedVectors[i].ux, vl->allocatedVectors[i].uy, 
	       vl->allocatedVectors[i].uz );
    }
  }
  fprintf( f, "end\n" );
  fprintf( f, "\n" );
    
  for ( i=0; i<cl->n_cones; i++ ) {

    if ( 0 ) 
      fprintf( f, "ic=cmin+round( (%f-emin)*(cmax-cmin)/(emax-emin) );\n",
	       cl->allocatedCones[i].error );
    fprintf( f, "ic=cmin+round( (log(%f)-log(emin))*(cmax-cmin)/(log(emax)-log(emin)) );\n",
	     cl->allocatedCones[i].error );
    if ( 0 ) 
      fprintf( stderr, "color %f -> %f\n",
	       1+(cl->allocatedCones[i].error-emin)*(64-1)/(emax-emin),
	       1+(log(cl->allocatedCones[i].error)-log(emin))*(64-1)/(log(emax)-log(emin)) );
  
    if ( 1 ) {
      x1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].x;
      y1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].y;
      z1 = vl->allocatedVectors[cl->allocatedCones[i].iv1].z;
      x2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].x;
      y2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].y;
      z2 = vl->allocatedVectors[cl->allocatedCones[i].iv2].z;
      x3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].x;
      y3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].y;
      z3 = vl->allocatedVectors[cl->allocatedCones[i].iv3].z;
      n1 = sqrt( x1*x1 + y1*y1 + z1*z1 );
      n2 = sqrt( x2*x2 + y2*y2 + z2*z2 );
      n3 = sqrt( x3*x3 + y3*y3 + z3*z3 );
    fprintf( f, "p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], col(ic,:) );\n",
	     x1/n1, x2/n2, x3/n3, x1/n1,
	     y1/n1, y2/n2, y3/n3, y1/n1,
	     z1/n1, z2/n2, z3/n3, z1/n1 );
    }
    else {
      fprintf( f, "p=patch( [%f %f %f %f], [%f %f %f %f], [%f %f %f %f], col(ic,:) );\n",
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].ux, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv2].ux, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv3].ux, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].ux,
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].uy, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv2].uy, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv3].uy, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].uy,
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].uz, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv2].uz, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv3].uz, 
	       vl->allocatedVectors[cl->allocatedCones[i].iv1].uz );
    }
     fprintf( f, "set(p, 'FaceAlpha', alpha );\n" );
     fprintf( f, "\n" );
  }
  fprintf( f, "\n" );
  fprintf( f, "if (1)\n" );
  fprintf( f, "  axis( [0 1 0 1 0 1] );\n" );
  fprintf( f, "else\n" );
  fprintf( f, "  axis( [0.4 1 0 0.8 0 0.6] );\n" );
  fprintf( f, "end\n" );
  fprintf( f, "\n" );
  fprintf( f, "%% axis equal;\n" );
  fprintf( f, "view( [110 20]);\n" );
  fprintf( f, "grid;\n" );
  fprintf( f, "\n" );
  fprintf( f, "if (0)\n" );
  fprintf( f, "  print( gcf, '-djpeg85', 'figure.jpg' );\n" );
  fprintf( f, "end\n" );
  fprintf( f, "\n" );
  fprintf( f, "hold off\n" );
}
