/*************************************************************************
 * chamfer.h - chamfer distance mask computation
 *
 * $$
 *
 * LICENSE:
 * GPL v3.0 (see gpl-3.0.txt for details)
 *
 * AUTHOR:
 * Gregoire Malandain (gregoire.malandain@inria.fr)
 * 
 * CREATION DATE: 
 * May, 10 2005
 *
 * ADDITIONS, CHANGES
 *
 */

#ifndef _chamfer_h_
#define _chamfer_h_

/* MISC
 */

typedef enum {
  _CONVEXITY_,
  _MAXIMALITY_
} CriteriumType;


extern void _set_chamfer_debug();

extern void _unset_chamfer_debug();

extern void _set_chamfer_trace();

extern void _unset_chamfer_trace();

extern void _set_chamfer_verbose( int v );

extern void _set_criterium_to_maximality();

extern void _set_criterium_to_convexity();


/* CONES
 */

typedef enum {
  _UNSET_,
  _INSIDE_,
  _OUT_IJ_,
  _OUT_JK_,
  _OUT_KI_,
} ConeType;

typedef struct Cone {

  int iv1;
  int iv2;
  int iv3;

  ConeType type;

  double error;

  int flag;

} Cone;

typedef struct ConeList {

  Cone *allocatedCones;
  int n_allocated_cones;
  int n_cones;

} ConeList;

extern void _init_cone_list( ConeList *cl );

extern void _free_cone_list( ConeList *cl );

extern void _sort_cones( Cone *cl, int left, int right );



/* VECTORS
 */

typedef struct Vector {

  int x;
  int y;
  int z;

  double vx;
  double vy;
  double vz;

  double vn;

  double ux;
  double uy;
  double uz;

  Cone **c;
  int n_cones;

  int w;
  int f;
  int l;

  int flag;
  
  int index;

} Vector;

typedef struct VectorList {

  Vector *allocatedVectors;
  int n_allocated_vectors;
  int n_vectors;

} VectorList;

extern void _init_vector_list( VectorList *vl );

extern void _free_vector_list( VectorList *vl );









/* RESULTS
 */
typedef struct {
  int *w;
  double error;
  double dmin;
  double dmax;
  double epsilon;
  double elapsed_time;
} Result;

typedef struct {
  double error_optimal;
  int n_results;
  int n_allocated_results;
  Result *allocatedResults;
} ResultList;

extern void _init_results_list( ResultList *rl );

extern void _free_results_list( ResultList *rl );

extern void _print_results_list( FILE *f, ResultList *rl,  VectorList *vl );

extern void _print_results_list_in_latex( FILE *f, ResultList *rl,  VectorList *vl );

extern void _add_result_to_list(  ResultList *rl,  VectorList *vl,
			   double err, double dmin, double dmax, double eps,
				  double time, double init_time );




/* VECTORS AND CONES
 */

extern int _check_3D_maximality( VectorList *vl, Cone *c );

extern int _check_3D_convexity( VectorList *vl, ConeList *cl, Cone *c );

extern void _print_2D_vector_list( FILE *f, VectorList *vl );

extern void _print_3D_vector_list( FILE *f, VectorList *vl );

extern void _print_cone_list( FILE *f, VectorList *vl, ConeList *cl );





/* 2D MASK
 */

extern void _init_2D_mask( VectorList *vl, ConeList *cl, double *s );

extern void _build_2D_mask( VectorList *vl, ConeList *cl, 
			    double *s, int size );

extern void _build_2D_mask_wrt_error( FILE *fout, VectorList *vl, ConeList *cl, 
				      double *s, int size, int n_vectors,
				      double error );

extern void _symmetrise_2D_mask( VectorList *vl, ConeList *cl, double *s );

extern void _compute_2D_weights( FILE *fout, ResultList *rl, 
				 VectorList *vl, ConeList *cl,
				 int f, int l );




/* 3D MASK
 */

extern void _init_3D_mask( VectorList *vl, ConeList *cl, double *s );

extern void _symmetrise_3D_mask( VectorList *vl, ConeList *cl, double *voxel_size,
			  int anisotropy_y, int anisotropy_z );

extern void _build_3D_mask( VectorList *vl, ConeList *cl, 
			    double *s, int size );

extern void _build_3D_mask_wrt_error( FILE *fout, VectorList *vl, ConeList *cl, 
				      double *s, int size, int n_vectors, 
				      double error );

extern void _build_3D_mask_3x3x5( VectorList *vl, ConeList *cl, 
				  double *s );

extern void _build_3D_mask_3x5x5( VectorList *vl, ConeList *cl, 
				  double *s );

extern void _build_3D_mask_5x5x5( VectorList *vl, ConeList *cl, 
				  double *s );

extern void _build_3D_mask_7x7x7( VectorList *vl, ConeList *cl, 
				  double *s );

extern void _compute_3D_weights( FILE *fout, ResultList *rl, 
				 VectorList *vl, ConeList *cl,
				 int f, int l );






/* MATLAB
 */

extern void _print_2D_mask_matlab( FILE *f, VectorList *vl, ConeList *cl );

extern void _print_3D_mask_matlab( FILE *f, VectorList *vl, ConeList *cl );


#endif
