/*************************************************************************
 * gener-mask.c - chamfer distance mask computation
 *
 * $$
 *
 * LICENSE:
 * GPL v3.0 (see gpl-3.0.txt for details)
 *
 * AUTHOR:
 * Gregoire Malandain (gregoire.malandain@inria.fr)
 * 
 * CREATION DATE: 
 * May, 10 2005
 *
 * ADDITIONS, CHANGES
 *
 */

#include <stdio.h>
#include <chamfer.h>
#include <string.h>





static char *options = "\
  [-dim %d] [-mask-size|-mask|-size %d] [-error %f] [-vectors %d]\n\
  [-prebuild|-pb %d] [-build-only|-bo] [-y %f] [-z %f]\n\
  [-f %d] [-l %d]\n\
  [-res %s] [-matlab %s] [-latex %s]\n\
  [-help|--help] [-examples|-ex]";

static char *details = "\n\
\n\
 =====  mask construction =====\n\
-dim %d : space dimension (3 or 2 (default))\n\
-mask-size|-mask|-size %d : mask size\n\
 3 (default) -> mask 3x3 or 3x3x3\n\
 5 ->  mask 5x5 or 5x5x5\n\
 etc.\n\
-vectors %d : maximal number of vectors in mask\n\
   if -mask-size is also specified (after in the command line)\n\
   the vectors are restricted in the given neighborhood\n\
-error %f : maximal sector error\n\
   0.01 means 1%\n\
   can be used in conjunction with -mask-size and -vectors\n\
-prebuild|-pb %d : prebuilt masks (assume dim=3)\n\
   admissible values:\n\
   3, 333 -> mask 3x3x3\n\
   5, 555 -> mask 5x5x5\n\
   7, 777 -> mask 7x7x7\n\
   335\n\
   355\n\
   this was implemented for comparison with (Fouard, IVC, 2005)\n\
-y %f : pixel or voxel dimension along Y\n\
-z %f : voxel dimension along z\n\
-build-only|-bo: do not compute weights\n\
\n\
 =====  weight computing  =====\n\
-f %d : first weight value (default=1)\n\
-l %d : last weight value (default=20)\n\
\n\
 ===== output controlling =====\n\
-res %s : specify output text file (default is stdout)\n\
-matlab %s : specify output matlab file\n\
   nice plot of mask errors\n\
-latex %s : specify output matlab file\n\
   latex table of computed weights\n\
\n\
 =====        misc        =====\n\
-help : print this help\n\
-examples|-ex : print some command lines examples\n\
\n";

static void usage( char *prg, char *str, int flag )
{ 
  fprintf( stderr, "\n" );
  fprintf( stderr, "%s:\n%s\n\n", prg, options );
  if ( flag ) 
    fprintf( stderr, "%s\n", details );
  if ( str != NULL ) 
    fprintf( stderr, "Error: %s\n", str );
}




const char *examples[] = {
"-dim 3 -size 5 -matlab ... -latex ...",
"-dim 3 -pb 777 -matlab ... -latex ...",
"-dim 3 -size 7 -matlab ... -latex ...",
"-dim 3 -size 3 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -pb 335 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -pb 355 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -size 5 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -vectors 10 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -vectors 16 -size 5 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -vectors 16 -y 1.2 -z 2 -matlab ... -latex ...",
"-dim 3 -vectors 19 -y 1.2 -z 2 -matlab ... -latex ...",
"" };

static void print_examples( char *prg )
{
  int i;
  for ( i=0; strlen( examples[i] ) > 0; i ++ ) 
    fprintf( stdout, "  %s %s\n", prg, examples[i] );
}

















typedef enum {
  _PREBUILD_,
  _SIZE_,
  _ERROR_
} BuildingCriteriumType;


int main( int argc, char *argv[] )
{
  int i;
  int status;
  char text[256];
  
  int first = 1;
  int last = 20;
  int verbose = 0;
  
  FILE *fout = stdout;
  char *nout = NULL;

  FILE *fmatlab = NULL;
  char *nmatlab = NULL;

  FILE *flatex = NULL;
  char *nlatex = NULL;

  int anisotropy_y = 0;
  int anisotropy_z = 0;

  int build_only = 0;
  int prebuild = 0;

  double voxel_size[3] = { 1.0, 1.0, 1.0 };
  int dim = 2;

  int mask_size = 3;
  double mask_error = 0.0001;
  int nb_vectors = 1000;

  ResultList rl;
  VectorList vl;
  ConeList cl;

  BuildingCriteriumType criterium = _SIZE_;



  /*----------------------------------------
    parsing args
    ----------------------------------------*/
  if ( argc == 1 ) {
    usage( argv[0], NULL, 0 );
    return( 0 );
  }
    
  for ( i=1; i<argc; i++ ) {

    if ( argv[i][0] == '-' ) {

      /* general stuff
       */

      if ( strcmp ( argv[i], "-help" ) == 0  || 
	   strcmp ( argv[i], "--help" ) == 0 ) {
	usage( argv[0], NULL, 1 );
	return( 0 );
      }

      if ( strcmp ( argv[i], "-examples" ) == 0  || 
	   strcmp ( argv[i], "-ex" ) == 0 ) {
	print_examples( argv[0] );
	return( 0 );
      }

      else if ( strcmp ( argv[i], "-verbose" ) == 0 || 
		(strcmp ( argv[i], "-v" ) == 0 && argv[i][2] == '\0') ) {
	i += 1;
	if ( i >= argc) {
	  _set_chamfer_verbose( 1 );
	  i -= 1;
	}
	else {
	  status = sscanf( argv[i],"%d", &verbose );
	  if ( status <= 0 ) {
	    _set_chamfer_verbose( 1 );
	    i -= 1;
	  }
	  else {
	    _set_chamfer_verbose( verbose );
	  }
	}
      }

      else if ( strcmp ( argv[i], "-trace" ) == 0  ) {
	_set_chamfer_trace();
      }
      
      else if ( strcmp ( argv[i], "-debug" ) == 0  ) {
	_set_chamfer_debug();
      }
      
      /* building masks
       */

      else if ( strcmp ( argv[i], "-dim" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -dim", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &dim );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -dim", 0 );
	  return( 1 );
	}
	if ( dim != 2 && dim != 3 ) {
	  usage( argv[0], "dimension not handled yet (-dim)", 0 );
	  return( 1 );
	}
	  
      }
      
      else if ( strcmp ( argv[i], "-mask-size" ) == 0 
		|| strcmp ( argv[i], "-mask" ) == 0
		|| strcmp ( argv[i], "-size" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -mask-size", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &mask_size );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -mask-size", 0 );
	  return( 1 );
	}
      }
      
      else if ( strcmp ( argv[i], "-vectors" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -vectors", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &nb_vectors );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -vectors", 0 );
	  return( 1 );
	}
	criterium = _ERROR_;
	mask_size = 101;
      }

      else if ( strcmp ( argv[i], "-error" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -error", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%lf", &mask_error );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -error", 0 );
	  return( 1 );
	}
	criterium = _ERROR_;
	mask_size = 101;
      }
      
      else if ( strcmp ( argv[i], "-prebuild" ) == 0 
		|| strcmp ( argv[i], "-pb" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -prebuild", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &prebuild );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -prebuild", 0 );
	  return( 1 );
	}
	switch ( prebuild ) {
	case 3 :
	case 333 :
	case 5 :
	case 555 :
	case 7 :
	case 777 :
	case 335 :
	case 355 :
	  break;
	default :
	  usage( argv[0], "unknown mask (-prebuild)", 0 );
	}
	criterium = _PREBUILD_;
	dim = 3;
      }
      
      else if ( strcmp ( argv[i], "-build-only" ) == 0 
		|| strcmp ( argv[i], "-bo" ) == 0 ) {
	build_only = 1;
      }
      
      else if ( strcmp ( argv[i], "-y" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -y", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%lf", &(voxel_size[1]) );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -y", 0 );
	  return( 1 );
	}
	anisotropy_y = 1;
      }
      
      else if ( strcmp ( argv[i], "-z" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -z", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%lf", &(voxel_size[2]) );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -z", 0 );
	  return( 1 );
	}
	anisotropy_z = 1;
      }

      else if ( strcmp ( argv[i], "-maximality" ) == 0 
		|| strcmp ( argv[i], "-max" ) == 0 ) {
	_set_criterium_to_maximality();
      }
      
      else if ( strcmp ( argv[i], "-convexity" ) == 0 
		|| strcmp ( argv[i], "-conv" ) == 0 ) {
	_set_criterium_to_convexity();
      }
      
      /* weights computation
       */
      
      else if ( strcmp ( argv[i], "-f" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -f", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &first );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -f", 0 );
	  return( 1 );
	}
      }
      
      else if ( strcmp ( argv[i], "-l" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -l", 0 );
	  return( 1 );
	}
	status = sscanf( argv[i],"%d", &last );
	if ( status <= 0 ) {
	  usage( argv[0], "error when parsing -l", 0 );
	  return( 1 );
	}
      }

      /* output parameters
       */
      
      else if ( strcmp ( argv[i], "-res" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -res", 0 );
	  return( 1 );
	}
	nout = argv[i];
      }
      
      else if ( strcmp ( argv[i], "-matlab" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -matlab", 0 );
	  return( 1 );
	}
	nmatlab = argv[i];
      }
      
      else if ( strcmp ( argv[i], "-latex" ) == 0 ) {
	i += 1;
	if ( i >= argc) {
	  usage( argv[0], "missing arg to -latex", 0 );
	  return( 1 );
	}
	nlatex = argv[i];
      }
      
      /* unknown options
       */

      else {
	sprintf(text,"unknown option %s\n",argv[i]);	
	usage( argv[0], text, 0 );
      }
    }

    else {
      sprintf(text,"unknown option %s\n",argv[i]);	
      usage( argv[0], text, 0 );
    }
  }





  /*----------------------------------------

    ----------------------------------------*/

  if ( nout != NULL ) {
    fout = fopen( nout, "w" );
    if ( fout == NULL ) {
      usage( argv[0], "error when opening result file\n", 0 );
      return( 1 );
    }
    fprintf( fout, "\n" );
    fprintf( fout, "%%" );
    for (i=0; i<argc; i++ )
      fprintf( fout, " %s", argv[i] );
    fprintf( fout, "\n\n" );
  }

  _init_results_list( &rl );
  _init_vector_list( &vl );
  _init_cone_list( &cl );


  
  /* Mask Initialization
   */
  switch( dim ) {
  case 2 :
  default :
    _init_2D_mask( &vl, &cl, voxel_size );
    break;
  case 3 :
    _init_3D_mask( &vl, &cl, voxel_size ); 
    break;
  }

  

  /* Symmetrisation, if required
   */
  switch( criterium ) {
  case _ERROR_ :
    switch( dim ) {
    case 2 :
    default :
      if ( anisotropy_y ) {
	_symmetrise_2D_mask( &vl, &cl, voxel_size );
      }
      break;
    case 3 :
      _symmetrise_3D_mask( &vl, &cl, voxel_size, anisotropy_y, anisotropy_z );
      break;
    }

    _sort_cones( cl.allocatedCones, 0, cl.n_cones-1 );
    break;
  default :
    break;
  }





  /* Mask Computation
   */
  switch( criterium ) {
    
  case _PREBUILD_ :
    switch ( prebuild ) {
    case 3 :
    case 333 :
      _build_3D_mask( &vl, &cl, voxel_size, mask_size );
      break;
    case 5 :
    case 555 :
      _build_3D_mask_5x5x5( &vl, &cl, voxel_size );
      break;
    case 7 :
    case 777 :
      _build_3D_mask_7x7x7( &vl, &cl, voxel_size );
      break;
    case 335 :
      _build_3D_mask_3x3x5( &vl, &cl, voxel_size );
      break;
    case 355 :
      _build_3D_mask_3x5x5( &vl, &cl, voxel_size );
      break;
      break;
    }
    break;

  case _SIZE_ :
    switch( dim ) {
    case 2 :
    default :
      _build_2D_mask( &vl, &cl, voxel_size, mask_size );
      break;
    case 3 :
      _build_3D_mask( &vl, &cl, voxel_size, mask_size );
      break;
    }
    break;

  case _ERROR_ :
    switch( dim ) {
    case 2 :
    default :
      _build_2D_mask_wrt_error( fout, &vl, &cl, voxel_size, 
				mask_size, nb_vectors, mask_error );
      break;
    case 3 :
      _build_3D_mask_wrt_error( fout, &vl, &cl, voxel_size, 
				mask_size,  nb_vectors, mask_error );
      break;
    }
    break;
  }


  /* Symmetrisation, if required
   */
  switch( criterium ) {
    
  case _PREBUILD_ :
    _symmetrise_3D_mask( &vl, &cl, voxel_size, anisotropy_y, anisotropy_z );
    _sort_cones( cl.allocatedCones, 0, cl.n_cones-1 );
    break;

  case _SIZE_ :
    switch( dim ) {
    case 2 :
    default :
      if ( anisotropy_y ) {
	_symmetrise_2D_mask( &vl, &cl, voxel_size );
      }
      break;
    case 3 :
      _symmetrise_3D_mask( &vl, &cl, voxel_size, anisotropy_y, anisotropy_z );
      break;
    }

    _sort_cones( cl.allocatedCones, 0, cl.n_cones-1 );
    break;
  default :
    break;
  }



  /* sectors are sorted
   */
  rl.error_optimal = cl.allocatedCones[0].error;



  /* Print Stuff
   */
  switch( dim ) {
  case 2 :
  default :
    _print_2D_vector_list( fout, &vl );
    _print_cone_list( fout, &vl, &cl );
    break;
  case 3 :
    _print_3D_vector_list( fout, &vl );
    _print_cone_list( fout, &vl, &cl );
    break;
  }


  if ( nmatlab != NULL ) {
    fmatlab = fopen( nmatlab, "w" );
    if ( fmatlab == NULL ) {
      usage( argv[0], "error when opening matlab file\n", 0 );
      return( 1 );
    }
    switch( dim ) {
    case 2 :
    default :
      _print_2D_mask_matlab( fmatlab, &vl, &cl );
      break;
    case 3 :
      _print_3D_mask_matlab( fmatlab, &vl, &cl );
      break;
    }
    fclose( fmatlab );
  }


  if ( build_only == 0 ) {

    /* compute weights
     */

    switch( dim ) {
    case 2 :
    default :
      _compute_2D_weights( fout, &rl, &vl, &cl, first, last );
      break;
    case 3 :
      _compute_3D_weights( fout, &rl, &vl, &cl, first, last );
      break;
    }
  }

  if ( rl.n_results > 0 ) {

    _print_results_list( fout, &rl, &vl );
    
    if ( nlatex != NULL ) {
      flatex = fopen( nlatex, "w" );
      if ( flatex == NULL ) {
	usage( argv[0], "error when opening latex file\n", 0 );
	return( 1 );
      }
      _print_results_list_in_latex( flatex, &rl, &vl );
      fclose( flatex );
    }

  }

  _free_results_list( &rl );
  _free_cone_list( &cl );
  _free_vector_list( &vl );

  if ( nout != NULL ) {
    fclose( fout );
  }

  return( 0 );
}
